/*
 * $QNXLicenseC:
 * Copyright 2012, QNX Software Systems. All Rights Reserved.
 *
 * You must obtain a written license from and pay applicable license fees to QNX
 * Software Systems before you may reproduce, modify or distribute this software,
 * or any work that includes all or part of this software.   Free development
 * licenses are available for evaluation and non-commercial purposes.  For more
 * information visit http://licensing.qnx.com or email licensing@qnx.com.
 *
 * This file may contain contributions from others.  Please review this entire
 * file for other proprietary rights or license notices, as well as the QNX
 * Development Suite License Guide at http://licensing.qnx.com/license-guide/
 * for other information.
 * $
 */
#ifdef HD_DEC_CLIENT
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <devctl.h>
#include <sys/types.h>
#include <sys/neutrino.h>
#include <sys/iomsg.h>
#include <sys/select.h>
#include <errno.h>
#include <pthread.h>

#include "hd-dec-client.h"

struct hd_dec_client *hd_dec = NULL;

static int
query_status(int fd, hd_dec_uint32_t *hd_dec_attr)
{
	int rc = EOK;

	if ((rc = devctl(fd, DCMD_HD_DEC_GET_CHANGES, hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
		fprintf(stderr, "devctl(%d) failed(%s)\n", DCMD_HD_DEC_GET_CHANGES, strerror(rc));
		return rc;
	}

	if(hd_dec_attr->property & HD_DEC_STATUS_A_SAMPLE_RATE) {
		if((rc = devctl(fd, DCMD_HD_DEC_GET_A_SAMPLE_RATE, hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
			fprintf(stderr, "devctl(%d) failed(%s)\n", DCMD_HD_DEC_GET_A_SAMPLE_RATE, strerror(rc));
			return rc;
		}
		if(hd_dec->verbosity) {
			fprintf(stderr, "sample rate changed: %d\n", hd_dec_attr->property);
		}
	}
	if(hd_dec_attr->property & HD_DEC_STATUS_A_MUTE) {
		if((rc = devctl(fd, DCMD_HD_DEC_GET_A_MUTE, hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
			fprintf(stderr, "devctl(%d) failed(%s)\n", DCMD_HD_DEC_GET_A_MUTE, strerror(rc));
			return rc;
		}
		if(hd_dec->verbosity) {
			fprintf(stderr, "audio mute state changed: %d\n", hd_dec_attr->property);

		}
	}
	return rc;
}

void
hd_dec_client_fini()
{
	if(hd_dec->fd != -1) {
		close(hd_dec->fd);
	}

	if(hd_dec != NULL) {
		free(hd_dec);
	}
}

int
hd_dec_client_init(int verbosity)
{
	uint32_t ndevices = -1;
	hd_dec_uint32_t hd_dec_attr;
	int i, rc = 0;

	if ((hd_dec = calloc(1, sizeof(struct hd_dec_client))) == NULL) {
		goto out;
	}

	hd_dec->fd = -1;
	hd_dec->verbosity = verbosity;

	if((hd_dec->fd = open(HD_DEC_MGR_NAME, O_RDWR, 0666)) == -1) {
		rc = errno;
		fprintf(stderr, "open %s failed(%s)\n", HD_DEC_MGR_NAME, strerror(rc));
		goto out;
	}

	if ((rc = devctl(hd_dec->fd, DCMD_HD_DEC_GET_NDEVICES, &ndevices, sizeof(uint32_t), NULL)) != EOK) {
		fprintf(stderr, "devctl(%d) failed(%s)\n", HD_DEC_GET_NDEVICES, strerror(rc));
		goto out;
	}
	if(hd_dec->verbosity) {
		printf("get ndevices = %d\n", ndevices);
	}

	for(i = 0; i < ndevices; i++) {
		hd_dec_attr.device_id = i;

		if ((rc = devctl(hd_dec->fd, DCMD_HD_DEC_GET_NSOURCES, &hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
			fprintf(stderr, "devctl(%d) failed(%s)\n", HD_DEC_GET_NSOURCES, strerror(rc));
			goto out;
		}
		if(hd_dec->verbosity) {
			printf("get nsources for device %d = %d\n", hd_dec_attr.device_id, hd_dec_attr.property);
		}
	}
out:
	return rc? -1: 0;
}

static void*
hd_dec_thread(void* arg)
{
	struct hd_dec_client *hd_dec = (struct hd_dec_client *)arg;
	hd_dec_uint32_t hd_dec_attr;
	sigset_t sigset;
	int rc;

	pthread_setname_np(pthread_self(), "hd_dec_thread");
	sigfillset(&sigset);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	fd_set  except_fd;
	FD_ZERO (&except_fd);
	FD_SET (hd_dec->fd, &except_fd);

	do {
		while((rc = select(1 + hd_dec->fd, 0, 0, &except_fd, NULL)) != -1) {
			if (rc > 0 && FD_ISSET (hd_dec->fd, &except_fd)) {
				hd_dec_attr.device_id = hd_dec->device;
				query_status(hd_dec->fd, &hd_dec_attr);
			}
		}
	} while (errno == EINTR);

	fprintf(stderr, "terminating hd_dec_thread as select() returns -1 (%s)\n", strerror(errno));
	return NULL;
}

static int
create_hd_dec_thread()
{
	pthread_attr_t tattr;
	int policy;
	struct sched_param parm;
	int err = 0;

	pthread_attr_init(&tattr);
	pthread_getschedparam(pthread_self(), &policy, &parm);
	pthread_attr_setinheritsched(&tattr, PTHREAD_EXPLICIT_SCHED);
	pthread_attr_setschedpolicy(&tattr, policy);
	pthread_attr_setschedparam(&tattr, &parm);

	err = pthread_create(NULL, &tattr, hd_dec_thread, hd_dec);

	(void)pthread_attr_destroy(&tattr);

	return err;
}

int
hd_dec_client_update(int device, int source, int nlanes, int verbosity)
{
	hd_dec_uint32_t hd_dec_attr;
	int rc;

	hd_dec->device = device;
	hd_dec->src_index = source;
	hd_dec->csi2_nlanes = nlanes;
	hd_dec->verbosity = verbosity;

	hd_dec_attr.device_id = hd_dec->device;
	hd_dec_attr.property = hd_dec->src_index;
	if ((rc = devctl(hd_dec->fd, DCMD_HD_DEC_SET_DEVICE_SOURCE, &hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
		fprintf(stderr, "devctl(%d) failed(%s)\n", HD_DEC_SET_DEVICE_SOURCE, strerror(rc));
		goto out;
	}
	if(hd_dec->verbosity) {
		printf("set device %d, source %d\n", hd_dec->device, hd_dec->src_index);
	}

	hd_dec_attr.property = hd_dec->csi2_nlanes;
	if ((rc = devctl(hd_dec->fd, DCMD_HD_DEC_SET_V_CSI2_NUM_DATA_LANES, &hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
		fprintf(stderr, "devctl(%d) failed(%s)\n", HD_DEC_SET_V_CSI2_NUM_DATA_LANES, strerror(rc));
		goto out;
	}
	if(hd_dec->verbosity) {
		printf("set %d CSI2 data lanes\n",hd_dec->csi2_nlanes);
	}

	if((rc = create_hd_dec_thread(hd_dec)) != EOK) {
		fprintf(stderr, "create_capture_thread failed:%s", strerror(rc));
		goto out;
	}
out:
	return rc? -1: 0;
}

int
hd_dec_v_output_enable(int enable)
{
	hd_dec_uint32_t hd_dec_attr;
	int rc;

	hd_dec->output_enable = enable;

	hd_dec_attr.device_id = hd_dec->device;
	hd_dec_attr.property = hd_dec->output_enable;
	if ((rc = devctl(hd_dec->fd, DCMD_HD_DEC_SET_V_OUTPUT_ENABLE, &hd_dec_attr, sizeof(hd_dec_uint32_t), NULL)) != EOK) {
		fprintf(stderr, "devctl(%d) failed(%s)\n", DCMD_HD_DEC_SET_V_OUTPUT_ENABLE, strerror(rc));
		return -1;
	}
	if(hd_dec->verbosity) {
		printf("enable output\n");
	}
	return 0;

}
#endif
#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn/product/branches/6.6.0/trunk/apps/screen/demos/vcapture-test/hd-dec-client.c $ $Rev: 730566 $")
#endif
