/*
 * $QNXLicenseC:
 * Copyright 2012, QNX Software Systems. All Rights Reserved.
 *
 * You must obtain a written license from and pay applicable license fees to QNX
 * Software Systems before you may reproduce, modify or distribute this software,
 * or any work that includes all or part of this software.   Free development
 * licenses are available for evaluation and non-commercial purposes.  For more
 * information visit http://licensing.qnx.com or email licensing@qnx.com.
 *
 * This file may contain contributions from others.  Please review this entire
 * file for other proprietary rights or license notices, as well as the QNX
 * Development Suite License Guide at http://licensing.qnx.com/license-guide/
 * for other information.
 * $
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/keycodes.h>
#include <time.h>
#include <screen/screen.h>
#include <errno.h>
#include <pthread.h>
#include <vcapture/capture.h>
#ifdef CAPTURE_J5
#include <vcapture/capture-j5-ext.h>
#endif
#ifdef CAPTURE_OMAP4_5
#include <vcapture/capture-omap4-5-ext.h>
#endif
#ifdef CAPTURE_ADV
#include <vcapture/capture-adv-ext.h>
#endif
#ifdef CAPTURE_IMX6X
#include <vcapture/capture-imx6x-ext.h>
#endif
#ifdef HD_DEC_CLIENT
#include "hd-dec-client.h"
#endif

static volatile int aborted = 0;
static const int exceptsigs[] = {
	SIGCHLD, SIGIO, SIGURG, SIGWINCH,
	SIGSTOP, SIGTSTP, SIGTTIN, SIGTTOU, SIGCONT,
	-1,
};

static void*
signal_thread(void* arg)
{
	sigset_t sigset;
	int sig;
	int i;

	pthread_setname_np(pthread_self(), "signal_thread");
	pthread_detach(pthread_self());
	sigfillset(&sigset);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	for (i = 0; exceptsigs[i] != -1; i ++) {
		sigdelset(&sigset, exceptsigs[i]);
	}

	for(;;) {
		if(sigwait(&sigset, &sig) == 0) {
			aborted = 1;
			break;
		}
	}
	return NULL;
}

void get_video_info(capture_context_t context)
{
	char *cur_norm = NULL;
	if(capture_is_property(context, CAPTURE_PROPERTY_CURRENT_NORM)) {
		capture_get_property_p(context, CAPTURE_PROPERTY_CURRENT_NORM, (void **)&cur_norm);
	}
	fprintf(stderr, "current norm: %s", cur_norm? cur_norm : "unavailable");
#ifdef CAPTURE_ADV
	int32_t lock = -1, fsclock = -1, freq = -1, wss = -1;
	if(capture_is_property(context, CAPTURE_PROPERTY_ADV_LOCK_STATUS)) {
		capture_get_property_i(context, CAPTURE_PROPERTY_ADV_LOCK_STATUS, &lock);
	}
	if(capture_is_property(context, CAPTURE_PROPERTY_ADV_FSCLOCK_STATUS)) {
		capture_get_property_i(context, CAPTURE_PROPERTY_ADV_FSCLOCK_STATUS, &fsclock);
	}
	if(capture_is_property(context, CAPTURE_PROPERTY_ADV_OUTPUT_FREQ)) {
		capture_get_property_i(context, CAPTURE_PROPERTY_ADV_OUTPUT_FREQ, &freq);
	}
	if(capture_is_property(context, CAPTURE_PROPERTY_ADV_WSS_STATUS)) {
		capture_get_property_i(context, CAPTURE_PROPERTY_ADV_WSS_STATUS, &wss);
	}
	fprintf(stderr, " lock:%d fsclock: %d freq: %d wss:%d", lock, fsclock, freq, wss);
#endif
	fprintf(stderr, "\n");
}

int change_color_settings(capture_context_t context, int color_test)
{
	static int contrast_val = -128;
	static int brightness_val = -128;
	static int saturation_val = -128;
	static int hue_val = -128;
	int rc;

	// contrast test
	if(color_test == 1 && capture_is_property( context, CAPTURE_PROPERTY_CONTRAST )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_CONTRAST, contrast_val += 1);
		if (rc) {
			if(contrast_val >= 127) {
				contrast_val = -128;
				printf("brightness test completes, restarting from -128\n");
			}else {
				perror("capture_set_property_i(CAPTURE_PROPERTY_CONTRAST)");
				return -1;
			}
		}
	}
	// brightness test
	if(color_test == 2 && capture_is_property( context, CAPTURE_PROPERTY_BRIGHTNESS )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_BRIGHTNESS, brightness_val += 1);
		if (rc) {
			if(brightness_val >= 127) {
				brightness_val = -128;
				fprintf(stderr, "brightness test completes, restarting from -128\n");
			}else {
				perror("capture_set_property_i(CAPTURE_PROPERTY_BRIGHTNESS)");
				return -1;
			}
		}
	}

	if(color_test == 3 && capture_is_property( context, CAPTURE_PROPERTY_SATURATION )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SATURATION, saturation_val += 1);
		if (rc) {
			if(saturation_val >= 127) {
				saturation_val = -128;
				fprintf(stderr, "saturation test completes, restarting from -128\n");
			}else {
				perror("capture_set_property_i(CAPTURE_PROPERTY_SATURATION)");
				return -1;
			}
		}
	}
	if(color_test == 4 && capture_is_property( context, CAPTURE_PROPERTY_HUE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_HUE, hue_val += 1);
		if (rc) {
			if(hue_val >= 127) {
				hue_val = -128;
				fprintf(stderr, "hue test completes, restarting from -128\n");
			}else {
				perror("capture_set_property_i(CAPTURE_PROPERTY_HUE)");
				return -1;
			}
		}
	}

	// commit changes to the H/W
	if(capture_update( context, 0 ) == -1) {
		fprintf(stderr, "capture_update failed\n");
		return -1;
	}
	return 0;
}

int main(int argc, char **argv)
{
	screen_context_t screen_ctx = NULL;
	screen_window_t screen_win = NULL;
	screen_display_t *screen_disp;	/* native handle for our display */
	int screen_ndisplays;		/* number of displays */
	const char *display = NULL;	/* the display to create our window on */
	int screen_val;

	int device = -1;
	int source = -1;
	int fid_pol = -1;
	int vsync_pol = -1;
	int hsync_pol = -1;
	int clock_pol = -1;
	int sfsize[2] = { -1, -1 };
	int dfsize[2] = { -1, -1};
	int bsize[2] = { -1, -1 };
	int size[2] = { -1, -1 };
	int pos[2] = { 0, 0 };
	int ssize[2] = {-1, -1};
	int spos[2] = { 0, 0 };
	int csize[2] = { -1, -1 };
	int cpos[2] = { 0, 0 };
	int format = SCREEN_FORMAT_YVYU;
	int pipeline = -1;
	int nbufs = -1;
	const char *tok;
	int rval = EXIT_FAILURE;
	int i, rc, val;
	capture_context_t context = NULL;

	int interlaced = SCREEN_INTERLACED_NONE;
	int n_pointers = 0;
	int buf_idx, prev_buf_idx = -1;
	int stride;
	int deinterlace_mode = CAPTURE_DEINTERLACE_NONE_MODE;
	int capture_priority = -1;
	int total_frame_cnt = 0x7FFFFFFF;
	int verbosity = 0;
	int brightness = 0xFFF;
	int contrast = 0xFFF;
	int saturation = 0xFFF;
	int hue = 0xFFF;
	int color_test = 0;
	int data_bus_width = 0;
	int video_info_frame_cnt = 0;
	char norm[100] = "";
	double frame_rate = -1.0;
	int interface_type = -1;
	int mem_fill = -1;
	bool quit_if_no_video = false;
	int zorder = 0;
	int nlanes = -1;
	bool use_resmgr = false;
	int reenable_delay_msec = -1;

#ifdef CAPTURE_J5
	int sync_type = -1;
	int actvid_hsync = -1;
	int edma_chan = -1;
	int adaptive_dei_int_mode = J5_ADAPTIVE_DEI_INT_MODE_3;
#endif
#ifdef CAPTURE_OMAP4_5
	int ddr_clk = -1;
#endif
#ifdef CAPTURE_ADV
	int source_type = -1;
#endif
#ifdef CAPTURE_IMX6X
	int sensor_clk_mode = -1;
	int adaptive_dei_motion_mode = VDI_MOTION_MODE_MED;
#endif
	sigset_t sigset;
	sigfillset(&sigset);
	for (i = 0; exceptsigs[i] != -1; i ++) {
		sigdelset(&sigset, exceptsigs[i]);
	}
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	for (i = 1; i < argc; i++) {
		if (!strncmp(argv[i], "-format=", strlen("-format="))) {
			tok = argv[i]+strlen("-format=");
			if (!strcmp(tok, "yvyu")) {
				format = SCREEN_FORMAT_YVYU;
			}else if (!strcmp(tok, "yuy2")) {
				format = SCREEN_FORMAT_YUY2;
                        }else if (!strcmp(tok, "uyvy")) {
				format = SCREEN_FORMAT_UYVY;
			}else if (!strcmp(tok, "rgb888")) {
				format = SCREEN_FORMAT_RGB888;
                        } else {
				fprintf(stderr, "invalid format: %s\n", argv[i] + strlen("-format="));
			}
		} else if (!strncmp(argv[i], "-device=", strlen("-device="))) {
			tok = argv[i] + strlen("-device=");
			device = atoi(tok);
		} else if (!strncmp(argv[i], "-source=", strlen("-source="))) {
			tok = argv[i] + strlen("-source=");
			source = atoi(tok);
		} else if (!strncmp(argv[i], "-fid-pol=", strlen("-fid-pol="))) {
			tok = argv[i] + strlen("-fid-pol=");
			fid_pol = atoi(tok);
		} else if (!strncmp(argv[i], "-vsync-pol=", strlen("-vsync-pol="))) {
			tok = argv[i] + strlen("-vsync-pol=");
			vsync_pol = atoi(tok);
		} else if (!strncmp(argv[i], "-hsync-pol=", strlen("-hsync-pol="))) {
			tok = argv[i] + strlen("-hsync-pol=");
			hsync_pol = atoi(tok);
		} else if (!strncmp(argv[i], "-clock-pol=", strlen("-clock-pol="))) {
			tok = argv[i] + strlen("-clock-pol=");
			clock_pol = atoi(tok);
		} else if (!strncmp(argv[i], "-dmode=", strlen("-dmode="))) {
			tok = argv[i]+strlen("-dmode=");
			if (!strcmp(tok, "bob")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_BOB_MODE;
			} else if(!strcmp(tok, "bob2")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_BOB2_MODE;
			} else if(!strcmp(tok, "weave")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_WEAVE_MODE;
			} else if(!strcmp(tok, "weave2")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_WEAVE2_MODE;
			} else if(!strcmp(tok, "adaptive")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_MOTION_ADAPTIVE_MODE;
			} else if(!strcmp(tok, "none")) {
				deinterlace_mode = CAPTURE_DEINTERLACE_NONE_MODE;
			}else {
				fprintf(stderr, "invalid de-interlace mode: %s\n", argv[i] + strlen("-dmode="));
			}
		} else if (strncmp(argv[i], "-size=", strlen("-size=")) == 0) {
			tok = argv[i] + strlen("-size=");
			size[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			size[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-ssize=", strlen("-ssize=")) == 0) {
			tok = argv[i] + strlen("-ssize=");
			ssize[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			ssize[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-csize=", strlen("-csize=")) == 0) {
			tok = argv[i] + strlen("-csize=");
			csize[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			csize[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-sfsize=", strlen("-sfsize=")) == 0) {
			tok = argv[i] + strlen("-sfsize=");
			sfsize[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			sfsize[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-bsize=", strlen("-bsize=")) == 0) {
			tok = argv[i] + strlen("-bsize=");
			bsize[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			bsize[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-pos=", strlen("-pos=")) == 0) {
			tok = argv[i] + strlen("-pos=");
			pos[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			pos[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-cpos=", strlen("-cpos=")) == 0) {
			tok = argv[i] + strlen("-cpos=");
			cpos[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			cpos[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-spos=", strlen("-spos=")) == 0) {
			tok = argv[i] + strlen("-spos=");
			spos[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			spos[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-display=", strlen("-display=")) == 0) {
			display = argv[i] + strlen("-display=");
		} else if (strncmp(argv[i], "-pipeline=", strlen("-pipeline=")) == 0) {
			tok = argv[i] + strlen("-pipeline=");
			pipeline = atoi(tok);
		} else if (strncmp(argv[i], "-prio=", strlen("-prio=")) == 0) {
			tok = argv[i] + strlen("-prio=");
			capture_priority = atoi(tok);
		} else if (strncmp(argv[i], "-frame-count=", strlen("-frame-count=")) == 0) {
			tok = argv[i] + strlen("-frame-count=");
			total_frame_cnt = atoi(tok);
		} else if (strncmp(argv[i], "-reenable-delay=", strlen("-reenable-delay=")) == 0) {
			tok = argv[i] + strlen("-reenable-delay=");
			reenable_delay_msec = atoi(tok);
		} else if (strncmp(argv[i], "-brightness=", strlen("-brightness=")) == 0) {
			tok = argv[i] + strlen("-brightness=");
			brightness = atoi(tok);
		} else if (strncmp(argv[i], "-contrast=", strlen("-contrast=")) == 0) {
			tok = argv[i] + strlen("-contrast=");
			contrast = atoi(tok);
		} else if (strncmp(argv[i], "-hue=", strlen("-hue=")) == 0) {
			tok = argv[i] + strlen("-hue=");
			hue = atoi(tok);
		} else if (strncmp(argv[i], "-saturation=", strlen("-saturation=")) == 0) {
			tok = argv[i] + strlen("-saturation=");
			saturation = atoi(tok);
		} else if (strncmp(argv[i], "-hue=", strlen("-hue=")) == 0) {
			tok = argv[i] + strlen("-hue=");
			hue = atoi(tok);
		} else if (strncmp(argv[i], "-color-test=", strlen("-color-test=")) == 0) {
			tok = argv[i] + strlen("-color-test=");
			color_test = atoi(tok);
		} else if (strncmp(argv[i], "-video-info=", strlen("-video-info=")) == 0) {
			tok = argv[i] + strlen("-video-info=");
			video_info_frame_cnt = atoi(tok);
		} else if (strncmp(argv[i], "-data-bus-width=", strlen("-data-bus-width=")) == 0) {
			tok = argv[i] + strlen("-data-bus-width=");
			data_bus_width = atoi(tok);
		} else if (strncmp(argv[i], "-norm=", strlen("-norm=")) == 0) {
			tok = argv[i] + strlen("-norm=");
			strcpy(norm, tok);
		} else if (strncmp(argv[i], "-frame-rate=", strlen("-frame-rate=")) == 0) {
			tok = argv[i] + strlen("-frame-rate=");
			frame_rate = atof(tok);
			if(errno == ERANGE || frame_rate <= 0.0) {
				fprintf(stderr, "invalid frame rate: %s \n", tok);
			}
		} else if (strncmp(argv[i], "-interface-type=", strlen("-interface-type=")) == 0) {
			tok = argv[i] + strlen("-interface-type=");
			if (!strcmp(tok, "parallel")) {
				interface_type = CAPTURE_IF_PARALLEL;
			} else if(!strcmp(tok, "csi2")) {
				interface_type = CAPTURE_IF_MIPI_CSI2;
			} else {
				fprintf(stderr, "invalid interface type: %s\n", argv[i] + strlen("-interface-type="));
			}
		} else if (strncmp(argv[i], "-verbosity=", strlen("-verbosity=")) == 0) {
			tok = argv[i] + strlen("-verbosity=");
			verbosity = atoi(tok);
		} else if (strncmp(argv[i], "-fill=", strlen("-fill=")) == 0) {
			tok = argv[i] + strlen("-fill=");
			mem_fill = atoi(tok);
		} else if (strncmp(argv[i], "-nbufs=", strlen("-nbufs=")) == 0) {
			tok = argv[i] + strlen("-nbufs=");
			nbufs = atoi(tok);
		} else if (strncmp(argv[i], "-nlanes=", strlen("-nlanes=")) == 0) {
			tok = argv[i] + strlen("-nlanes=");
			nlanes = atoi(tok);
		} else if (strncmp(argv[i], "-quit-if-no-video", strlen("-quit-if-no-video")) == 0) {
			quit_if_no_video = true;
		} else if (strncmp(argv[i], "-resmgr", strlen("-resmgr")) == 0) {
			use_resmgr = true;
		} else if (strncmp(argv[i], "-zorder=", strlen("-zorder=")) == 0) {
			tok = argv[i] + strlen("-zorder=");
			zorder = atoi(tok);
#ifdef CAPTURE_J5
		} else if (!strncmp(argv[i], "-sync-type=", strlen("-sync-type="))) {
			tok = argv[i] + strlen("-sync-type=");
			sync_type = atoi(tok);
		} else if (!strncmp(argv[i], "-actvid-hsync=", strlen("-actvid-hsync="))) {
			tok = argv[i] + strlen("-actvid-hsync=");
			actvid_hsync = atoi(tok);
		} else if (!strncmp(argv[i], "-adaptive-dei-int-mode=", strlen("-adaptive-dei-int-mode="))) {
			tok = argv[i]+strlen("-adaptive-dei-int-mode=");
			if ((!strcmp(tok, "mode0")) || (!strcmp(tok, "0"))) {
				adaptive_dei_int_mode = J5_ADAPTIVE_DEI_INT_MODE_0;
			} else if((!strcmp(tok, "mode1")) || (!strcmp(tok, "1"))) {
				adaptive_dei_int_mode = J5_ADAPTIVE_DEI_INT_MODE_1;
			} else if((!strcmp(tok, "mode2")) || (!strcmp(tok, "2"))) {
				adaptive_dei_int_mode = J5_ADAPTIVE_DEI_INT_MODE_2;
			} else if((!strcmp(tok, "mode3")) || (!strcmp(tok, "3"))) {
				adaptive_dei_int_mode = J5_ADAPTIVE_DEI_INT_MODE_3;
			} else {
				fprintf(stderr, "invalid adaptive DEI interpolation mode: %s.\n"
						"Using mode 3: Edge-directed interpolation for Luma & Chroma, using "
						"MV from MDT.\n", argv[i] + strlen("-adaptive-dei-int-mode="));
			}
		} else if (strncmp(argv[i], "-edma-chan=", strlen("-edma-chan=")) == 0) {
			tok = argv[i] + strlen("-edma-chan=");
			edma_chan = atoi(tok);
#endif
#ifdef CAPTURE_OMAP4_5
		} else if (strncmp(argv[i], "-ddr-clk=", strlen("-ddr-clk=")) == 0) {
			tok = argv[i] + strlen("-ddr-clk=");
			ddr_clk = atoi(tok);
#endif
#ifdef CAPTURE_ADV
		} else if (strncmp(argv[i], "-source-type=", strlen("-source-type=")) == 0) {
			tok = argv[i] + strlen("-source-type=");
			source_type = atoi(tok);
#endif
#ifdef CAPTURE_IMX6X
		} else if (strncmp(argv[i], "-sensor-clk-mode=", strlen("-sensor-clk-mode=")) == 0) {
			tok = argv[i] + strlen("-sensor-clk-mode=");
			sensor_clk_mode = atoi(tok);
		} else if (!strncmp(argv[i], "-adaptive-dei-motion-mode=", strlen("-adaptive-dei-motion-mode="))) {
			tok = argv[i]+strlen("-adaptive-dei-motion-mode=");
			if (!stricmp(tok, "low")) {
				adaptive_dei_motion_mode = VDI_MOTION_MODE_LOW;
			} else if(!stricmp(tok, "med")) {
				adaptive_dei_motion_mode = VDI_MOTION_MODE_MED;
			} else if(!stricmp(tok, "high")) {
				adaptive_dei_motion_mode = VDI_MOTION_MODE_HIGH;
			} else {
				fprintf(stderr, "invalid adaptive DEI motion mode: %s.\n"
						"Using medium motion mode\n", tok);
			}
#endif
		} else {
			fprintf(stderr, "invalid command line option: %s\n", argv[i]);
		}
	}

	if( nbufs == -1 ) {
		nbufs = deinterlace_mode == CAPTURE_DEINTERLACE_MOTION_ADAPTIVE_MODE ? 10 : 4;
	} else if( verbosity ) {
		fprintf(stderr, "application will attempt to create %d buffers\n", nbufs);
	}

	screen_buffer_t screen_buf[nbufs];
	void *pointers[nbufs];
	memset( pointers, 0, nbufs * sizeof(void *));

	uint64_t frame_timestamp[nbufs];
	memset( frame_timestamp, 0, nbufs * sizeof(uint64_t));

#ifdef CAPTURE_NEED_SEQNO
	uint32_t frame_seqno[nbufs];
	memset( frame_seqno, 0, nbufs * sizeof(uint32_t));
#endif
	uint32_t frame_flags[nbufs];
	memset( frame_flags, 0, nbufs * sizeof(uint32_t));

	if(bsize[0] == -1 || bsize[1] == -1) {
		fprintf(stderr, "error: buffer is not specified!\n");
		goto fail;
	}

	rc = pthread_create(NULL, NULL, signal_thread, NULL);
	if (rc) {
		perror("pthread_create failed");
		goto fail;
	}
#ifdef HD_DEC_CLIENT
	if(use_resmgr) {
		rc = hd_dec_client_init(verbosity);
		if(rc) {
			perror("hd_dec_client_init");
			goto fail;
		}
	}
#endif
	rc = screen_create_context(&screen_ctx, 0);
	if (rc) {
		perror("screen_context_create");
		goto fail;
	}

	rc = screen_create_window(&screen_win, screen_ctx);
	if (rc) {
		perror("screen_create_window");
		goto fail;
	}

	if (display) {
		rc = screen_get_context_property_iv(screen_ctx, SCREEN_PROPERTY_DISPLAY_COUNT,
			&screen_ndisplays);
		if (rc) {
			perror("screen_get_context_property_iv(SCREEN_PROPERTY_DISPLAY_COUNT)");
			goto fail;
		}

		screen_disp = calloc(screen_ndisplays, sizeof(*screen_disp));
		if (screen_disp == NULL) {
			fprintf(stderr, "could not allocate memory for display list\n");
			goto fail;
		}

		rc = screen_get_context_property_pv(screen_ctx, SCREEN_PROPERTY_DISPLAYS,
			(void **)screen_disp);
		if (rc) {
			perror("screen_get_context_property_ptr(SCREEN_PROPERTY_DISPLAYS)");
			free(screen_disp);
			goto fail;
		}

		if (isdigit(*display)) {
			int want_id = atoi(display);
			for (i = 0; i < screen_ndisplays; ++i) {
				int actual_id = 0;  // invalid
				(void)screen_get_display_property_iv(screen_disp[i],
						SCREEN_PROPERTY_ID, &actual_id);
				if (want_id == actual_id) {
					break;
				}
			}
		} else {
			int type = -1;
			if (strcmp(display, "internal") == 0) {
				type = SCREEN_DISPLAY_TYPE_INTERNAL;
			} else if (strcmp(display, "composite") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPOSITE;
			} else if (strcmp(display, "svideo") == 0) {
				type = SCREEN_DISPLAY_TYPE_SVIDEO;
			} else if (strcmp(display, "YPbPr") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_YPbPr;
			} else if (strcmp(display, "rgb") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_RGB;
			} else if (strcmp(display, "rgbhv") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_RGBHV;
			} else if (strcmp(display, "dvi") == 0) {
				type = SCREEN_DISPLAY_TYPE_DVI;
			} else if (strcmp(display, "hdmi") == 0) {
				type = SCREEN_DISPLAY_TYPE_HDMI;
			} else if (strcmp(display, "other") == 0) {
				type = SCREEN_DISPLAY_TYPE_OTHER;
			} else {
				fprintf(stderr, "unknown display type %s\n", display);
				free(screen_disp);
				goto fail;
			}
			for (i = 0; i < screen_ndisplays; i++) {
				screen_get_display_property_iv(screen_disp[i],
					SCREEN_PROPERTY_TYPE, &screen_val);
				if (screen_val == type) {
					break;
				}
			}
		}
		if (i >= screen_ndisplays) {
			fprintf(stderr, "couldn't find display %s\n", display);
			free(screen_disp);
			goto fail;
		}

		rc = screen_set_window_property_pv(screen_win, SCREEN_PROPERTY_DISPLAY,
		(void **)&screen_disp[i]);
			if (rc) {
			perror("screen_set_window_property_ptr(SCREEN_PROPERTY_DISPLAY)");
			free(screen_disp);
			goto fail;
		}

		free(screen_disp);
	}

	val = SCREEN_USAGE_WRITE|SCREEN_USAGE_VIDEO|SCREEN_USAGE_CAPTURE;
	if (pipeline != -1) {
		val |= SCREEN_USAGE_OVERLAY;
	}
	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_USAGE, &val);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_USAGE)");
		goto fail;
	}

	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_FORMAT, &format);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_FORMAT)");
		goto fail;
	}

	if (size[0] == -1 || size[1] == -1) {
		rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_SIZE, size);
		if (rc) {
			perror("screen_get_window_property_iv(SCREEN_PROPERTY_SIZE)");
			goto fail;
		}
	}

	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_BUFFER_SIZE, bsize);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_BUFFER_SIZE)");
		goto fail;
	}

	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SIZE, size);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_SIZE)");
		goto fail;
	}
	if (pos[0] != 0 || pos[1] != 0) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_POSITION, pos);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_POSITION)");
			goto fail;
		}
	}

	if (pipeline != -1) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_PIPELINE, &pipeline);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_PIPELINE)");
			goto fail;
		}
	}
	
	if (zorder){
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_ZORDER, &zorder);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_ZORDER)");
			goto fail;
		}
	}

	if (zorder){
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_ZORDER, &zorder);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_ZORDER)");
			goto fail;
		}
	}

	if (zorder){
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_ZORDER, &zorder);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_ZORDER)");
			goto fail;
		}
	}

	// Set ID string for debugging via /dev/screen.
	{
		const char *idstr = "vcapture-test";
		screen_set_window_property_cv(screen_win,
				SCREEN_PROPERTY_ID_STRING, strlen(idstr), idstr);
	}

	rc = screen_create_window_buffers(screen_win, nbufs);
	if (rc) {
		perror("screen_create_window_buffers");
		goto fail;
	}

	rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_RENDER_BUFFER_COUNT, &n_pointers);
	if (rc) {
		perror("screen_get_window_property_iv(SCREEN_PROPERTY_RENDER_BUFFER_COUNT)");
		goto fail;
	}

	rc = screen_get_window_property_pv(screen_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)screen_buf);
	if (rc) {
		perror("screen_get_window_property_pv(SCREEN_PROPERTY_RENDER_BUFFERS)");
		goto fail;
	}

	for (i = 0; i < n_pointers; i++) {
		// obtain the pointer of the buffers, for the capture use
		rc = screen_get_buffer_property_pv(screen_buf[i], SCREEN_PROPERTY_POINTER, &(pointers[i]));
		if (rc) {
			perror("screen_get_window_property_pv(SCREEN_PROPERTY_POINTER)");
			goto fail;
		}
	}

	rc = screen_get_buffer_property_iv(screen_buf[0], SCREEN_PROPERTY_STRIDE, &stride);
	if (rc) {
		perror("screen_get_buffer_property_iv(SCREEN_PROPERTY_STRIDE)");
		goto fail;
	}

	if(mem_fill >= 0) {
		for (i = 0; i < n_pointers; i++) {
			memset(pointers[i], mem_fill, stride * bsize[1]);
		}
	}

	// Connect to a capture device
	context = capture_create_context( 0 );
	if( !context ) {
		perror("capture_create_context failed");
		goto fail;
	}

	int ndevices;
	// get the number of devices controlled by capture libraries.
	rc = capture_get_property_i( context, CAPTURE_PROPERTY_NDEVICES, &ndevices );
	if (rc) {
		perror("capture_get_property_i(CAPTURE_PROPERTY_NDEVICES)");
		goto fail;
	}

	if (device >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DEVICE, device );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DEVICE)");
			goto fail;
		}
	}

	const char *info = NULL;
	capture_get_property_p( context, CAPTURE_PROPERTY_DEVICE_INFO, (void **)&info );

	// Validate device's properties
	if( !capture_is_property( context, CAPTURE_ENABLE )
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_DEINTERLACE_MODE )
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_CROP_WIDTH )
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_CROP_HEIGHT)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_CROP_X)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_CROP_Y)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_DST_WIDTH)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_DST_HEIGHT)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_DST_STRIDE)
	//	 || !capture_is_property( context, CAPTURE_PROPERTY_DST_NBYTES)
		 || !capture_is_property( context, CAPTURE_PROPERTY_FRAME_NBUFFERS)
		 || !capture_is_property( context, CAPTURE_PROPERTY_FRAME_BUFFERS)
		 || !capture_is_property( context, CAPTURE_PROPERTY_DST_FORMAT))
	 {
		fprintf( stderr, "can't do it. some properties are not supported by this driver\n" );
		goto fail;
	 }

#ifdef CAPTURE_ADV
	// note: for ADVxxxx decoder, the source type should be set, prior to the source index
	if (source_type >= 0 && capture_is_property( context, CAPTURE_PROPERTY_ADV_SRC_TYPE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_ADV_SRC_TYPE, source_type );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_ADV_SRC_TYPE)");
			goto fail;
		}
	}
#endif
	if (source >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SRC_INDEX, source );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_SRC_INDEX)");
			goto fail;
		}
	}
	if (fid_pol >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_INVERT_FID_POL, fid_pol );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_INVERT_FID_POL)");
			goto fail;
		}
	}

	if (vsync_pol >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_INVERT_VSYNC_POL, vsync_pol );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_INVERT_VSYNC_POL)");
			goto fail;
		}
	}

	if (hsync_pol >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_INVERT_HSYNC_POL, hsync_pol );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_INVERT_HSYNC_POL)");
			goto fail;
		}
	}

	if (clock_pol >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_INVERT_CLOCK_POL, clock_pol );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_INVERT_CLOCK_POL)");
			goto fail;
		}
	}
	if(capture_is_property( context, CAPTURE_PROPERTY_DEINTERLACE_MODE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DEINTERLACE_MODE, deinterlace_mode);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DEINTERLACE_MODE)");
			goto fail;
		}
		if(deinterlace_mode == CAPTURE_DEINTERLACE_WEAVE2_MODE) {
#ifdef CAPTURE_J5
			rc = capture_set_property_i( context, CAPTURE_PROPERTY_J5_EDMA_CHANNEL, edma_chan);
			if (rc) {
				perror("capture_set_property_i(CAPTURE_PROPERTY_J5_EDMA_CHANNEL)");
				goto fail;
			}
#endif
		}
		if (deinterlace_mode == CAPTURE_DEINTERLACE_MOTION_ADAPTIVE_MODE) {
#ifdef CAPTURE_J5
			if(capture_is_property( context, CAPTURE_PROPERTY_J5_ADAPTIVE_DEI_INT_MODE )) {
				rc = capture_set_property_i( context, CAPTURE_PROPERTY_J5_ADAPTIVE_DEI_INT_MODE, adaptive_dei_int_mode );
				if (rc) {
					perror("capture_set_property_i(CAPTURE_PROPERTY_J5_ADAPTIVE_DEI_INTERPOLATION_MODE)");
					goto fail;
				}
			}
#endif
#ifdef CAPTURE_IMX6X
			if(capture_is_property( context, CAPTURE_PROPERTY_IMX6X_ADAPTIVE_DEI_MOTION_MODE )) {
				rc = capture_set_property_i( context, CAPTURE_PROPERTY_IMX6X_ADAPTIVE_DEI_MOTION_MODE, adaptive_dei_motion_mode );
				if (rc) {
					perror("capture_set_property_i(CAPTURE_PROPERTY_IMX6X_ADAPTIVE_DEI_MOTION_MODE)");
					goto fail;
				}
			}
#endif
		}
	}
#ifdef CAPTURE_IMX6X
	if (sensor_clk_mode >= 0 && capture_is_property( context, CAPTURE_PROPERTY_IMX6X_SENSOR_CLK_MODE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_IMX6X_SENSOR_CLK_MODE, sensor_clk_mode );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_IMX6X_SENSOR_CLK_MODE)");
			goto fail;
		}
	}
#endif
	rc = capture_set_property_i( context, CAPTURE_PROPERTY_FRAME_NBUFFERS, n_pointers );
	if (rc) {
		perror("capture_set_property_i(CAPTURE_PROPERTY_FRAME_NBUFFERS)");
		goto fail;
	}
	rc = capture_set_property_p( context, CAPTURE_PROPERTY_FRAME_BUFFERS, pointers );
	if (rc) {
		perror("capture_set_property_p(CAPTURE_PROPERTY_FRAME_BUFFERS)");
		goto fail;
	}
	if(capture_is_property( context, CAPTURE_PROPERTY_DST_STRIDE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DST_STRIDE, stride);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DST_STRIDE)");
			goto fail;
		}
	}
	rc = capture_set_property_i( context, CAPTURE_PROPERTY_DST_FORMAT, format);
	if (rc) {
		perror("capture_set_property_i(CAPTURE_PROPERTY_DST_FORMAT)");
		goto fail;
	}

	if(data_bus_width > 0 && capture_is_property( context, CAPTURE_PROPERTY_DATA_BUS_WIDTH )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DATA_BUS_WIDTH, data_bus_width);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DATA_BUS_WIDTH)");
			goto fail;
		}
	}

	if(capture_is_property( context, CAPTURE_PROPERTY_SRC_WIDTH )
			&& capture_is_property( context, CAPTURE_PROPERTY_SRC_HEIGHT )
			&& (sfsize[0] != -1) && (sfsize[1] != -1)) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SRC_WIDTH, sfsize[0]);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_SRC_WIDTH)");
			goto fail;
		}
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SRC_HEIGHT, sfsize[1]);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_SRC_HEIGHT)");
			goto fail;
		}
	}

	dfsize[0] = bsize[0];
	dfsize[1] = bsize[1];

	if(capture_is_property( context, CAPTURE_PROPERTY_CROP_WIDTH )
			&& capture_is_property( context, CAPTURE_PROPERTY_CROP_HEIGHT )
			&& capture_is_property( context, CAPTURE_PROPERTY_CROP_X )
			&& capture_is_property( context, CAPTURE_PROPERTY_CROP_Y )) {

		if(csize[0] != -1 && csize[1] != -1) {
			rc = capture_set_property_i( context, CAPTURE_PROPERTY_CROP_WIDTH, csize[0]);
			if (rc) {
				perror("capture_set_property_i(CAPTURE_PROPERTY_CROP_WIDTH)");
				goto fail;
			}
			rc = capture_set_property_i( context, CAPTURE_PROPERTY_CROP_HEIGHT, csize[1]);
			if (rc) {
				perror("capture_set_property_i(CAPTURE_PROPERTY_CROP_HEIGHT)");
				goto fail;
			}
			// the scaling is not required on the capture side
			dfsize[0] = csize[0];
			dfsize[1] = csize[1];
		}
		if(cpos[0] != 0 || cpos[1] != 0) {
			rc = capture_set_property_i( context, CAPTURE_PROPERTY_CROP_X, cpos[0]);
			if (rc) {
				perror("capture_set_property_i(CAPTURE_PROPERTY_CROP_X)");
				goto fail;
			}
			rc = capture_set_property_i( context, CAPTURE_PROPERTY_CROP_Y, cpos[1]);
			if (rc) {
				perror("capture_set_property_i(CAPTURE_PROPERTY_CROP_Y)");
				goto fail;
			}
		}
	}

	if(capture_is_property( context, CAPTURE_PROPERTY_DST_WIDTH )
			&& capture_is_property( context, CAPTURE_PROPERTY_DST_HEIGHT )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DST_WIDTH, dfsize[0]);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DST_WIDTH)");
			goto fail;
		}
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_DST_HEIGHT, dfsize[1]);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_DST_HEIGHT)");
			goto fail;
		}
	}


	if(ssize[0] == -1 || ssize[1] == -1 || (ssize[0] + spos[0]) > dfsize[0] || (ssize[1] + spos[1]) > dfsize[1]) {
		fprintf(stderr, "adjusting the source viewport size from %d x %d ", ssize[0], ssize[1]);
		ssize[0] = dfsize[0] - spos[0];
		ssize[1] = dfsize[1] - spos[1];
		fprintf(stderr, "to %d x %d\n", ssize[0], ssize[1]);
	}

	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SOURCE_SIZE, ssize);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_SOURCE_SIZE)");
		goto fail;
	}

	if (spos[0] != 0 || spos[1] != 0) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SOURCE_POSITION, spos);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_SOURCE_POSITION)");
			goto fail;
		}
	}

	// tell the capture driver to use this array and update it when frames are captured.
	capture_set_property_p( context, CAPTURE_PROPERTY_FRAME_TIMESTAMP, frame_timestamp );
#ifdef CAPTURE_NEED_SEQNO
	capture_set_property_p( context, CAPTURE_PROPERTY_FRAME_SEQNO, frame_seqno );
#endif
	if(capture_is_property( context, CAPTURE_PROPERTY_FRAME_FLAGS )) {
		rc = capture_set_property_p( context, CAPTURE_PROPERTY_FRAME_FLAGS, frame_flags );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_FRAME_FLAGS)");
			goto fail;
		}
	}
	if(capture_priority != -1) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_THREAD_PRIORITY, capture_priority);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_THREAD_PRIORITY)");
			goto fail;
		}
	}
	if(brightness != 0xFFF && capture_is_property( context, CAPTURE_PROPERTY_BRIGHTNESS )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_BRIGHTNESS, brightness);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_BRIGHTNESS)");
			goto fail;
		}
	}

	if(contrast != 0xFFF && capture_is_property( context, CAPTURE_PROPERTY_CONTRAST )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_CONTRAST, contrast);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_CONTRAST)");
			goto fail;
		}
	}

	if(saturation != 0xFFF && capture_is_property( context, CAPTURE_PROPERTY_SATURATION )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SATURATION, saturation);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_SATURATION)");
			goto fail;
		}
	}
	if(hue != 0xFFF && capture_is_property( context, CAPTURE_PROPERTY_HUE )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_SATURATION, hue);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_HUE)");
			goto fail;
		}
	}

	if((strlen(norm) != 0) && (capture_is_property( context, CAPTURE_PROPERTY_NORM ))) {
		rc = capture_set_property_p(context, CAPTURE_PROPERTY_NORM, norm);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_NORM)");
			goto fail;
		}
	}

	if(interface_type != -1 && capture_is_property( context, CAPTURE_PROPERTY_INTERFACE_TYPE )) {
		rc = capture_set_property_i(context, CAPTURE_PROPERTY_INTERFACE_TYPE, interface_type);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_INTERFACE_TYPE)");
			goto fail;
		}
	}

	if(capture_is_property( context, CAPTURE_PROPERTY_CSI2_NUM_DATA_LANES )) {
		rc = capture_set_property_i(context, CAPTURE_PROPERTY_CSI2_NUM_DATA_LANES, nlanes);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_CSI2_NUM_DATA_LANES)");
			goto fail;
		}
	}

	if(capture_is_property( context, CAPTURE_PROPERTY_VERBOSITY )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_VERBOSITY, verbosity);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_VERBOSITY)");
			goto fail;
		}
	}

#ifdef CAPTURE_J5
	if (sync_type >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_J5_SYNC_TYPE, sync_type );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_J5_SYNC_TYPE)");
			goto fail;
		}
	}

	if (actvid_hsync >= 0) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_J5_USE_ACTVID_HSYNC, actvid_hsync );
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_J5_USE_ACTVID_HSYNC)");
			goto fail;
		}
	}
#endif
#ifdef CAPTURE_OMAP4_5
	if(capture_is_property( context, CAPTURE_PROPERTY_OMAP_DDR_CLOCK_MHZ )) {
		rc = capture_set_property_i( context, CAPTURE_PROPERTY_OMAP_DDR_CLOCK_MHZ, ddr_clk);
		if (rc) {
			perror("capture_set_property_i(CAPTURE_PROPERTY_OMAP_DDR_CLOCK_MHZ)");
			goto fail;
		}
	}
#endif

	// tell the driver to start capturing (when capture_update() is called).
	capture_set_property_i( context, CAPTURE_ENABLE, 1 );
#ifdef HD_DEC_CLIENT
	if(use_resmgr) {
		rc = hd_dec_client_update((device - ndevices), source, nlanes, verbosity);
		if(rc) {
			perror("hd_dec_client_update");
			goto fail;
		}
	}
#endif
	// commit changes to the H/W -- and start capturing...
	if(capture_update( context, 0 ) == -1) {
		fprintf(stderr, "capture_update failed\n");
		goto fail;
	}

#ifdef HD_DEC_CLIENT
	if(use_resmgr) {
		rc = hd_dec_v_output_enable(1);
		if(rc) {
			perror("hd_dec_v_output_enable");
			goto fail;
		}
	}
#endif
	int frame_no = 0;
	struct timespec ts;
	uint64_t t = UINT64_MAX;

	while(!aborted) {
		if(quit_if_no_video) {
			char *cur_norm = NULL;
			if(capture_is_property(context, CAPTURE_PROPERTY_CURRENT_NORM)) {
				capture_get_property_p(context, CAPTURE_PROPERTY_CURRENT_NORM, (void **)&cur_norm);
				if(cur_norm && (strcmpi(cur_norm, CAPTURE_NORM_NONE) == 0)) {
					sleep(1);
					// It takes decoders some time to detect whether the input video exists or not.
					// Check the input status again after 1 second, to make sure the video input is really not detected.
					capture_get_property_p(context, CAPTURE_PROPERTY_CURRENT_NORM, (void **)&cur_norm);
					if(strcmpi(cur_norm, CAPTURE_NORM_NONE) == 0) {
						fprintf(stderr, "no video input.\n");
						goto end;
					}
				}
			}
		}

		if(quit_if_no_video) {
			char *cur_norm = NULL;
			if(capture_is_property(context, CAPTURE_PROPERTY_CURRENT_NORM)) {
				capture_get_property_p(context, CAPTURE_PROPERTY_CURRENT_NORM, (void **)&cur_norm);
				if(cur_norm && (strcmpi(cur_norm, CAPTURE_NORM_NONE) == 0)) {
					sleep(1);
					// It takes decoders some time to detect whether the input video exists or not.
					// Check the input status again after 1 second, to make sure the video input is really not detected.
					capture_get_property_p(context, CAPTURE_PROPERTY_CURRENT_NORM, (void **)&cur_norm);
					if(strcmpi(cur_norm, CAPTURE_NORM_NONE) == 0) {
						fprintf(stderr, "no video input.\n");
						goto end;
					}
				}
			}
		}

		// get next captured frame...
		if(prev_buf_idx == -1) {  // first frame
			clock_gettime( CLOCK_REALTIME, &ts );
			t = timespec2nsec(&ts);
		}

		buf_idx = capture_get_frame( context, (frame_rate > 0.0)?
				(uint64_t)(CAPTURE_INTERVAL_FROM_MS(1000 / frame_rate)) : CAPTURE_TIMEOUT_INFINITE, 0 );
		if(buf_idx == -1) {
			fprintf(stderr, "capture_get_frame failed(%s)\n", strerror(errno));
			goto fail;
		}

		// The returned idx-ed pointer is 'locked' upon return from the capture_get_frame(), and
		// this buffer will remain locked until the capture_release_frame() is called with the corresponding index.
		// It is safe to access the contents of the frame_seqno[buf_idx], frame_timestamp[buf_idx], etc.
		if(frame_rate > 0.0) {
			if(prev_buf_idx == -1) { // first frame
				clock_gettime( CLOCK_REALTIME, &ts );
				t = timespec2nsec(&ts)-t;
			}
			else {
				t = frame_timestamp[buf_idx] - frame_timestamp[prev_buf_idx];
			}
			if( t > CAPTURE_INTERVAL_FROM_MS(1000 / frame_rate)) {
				fprintf( stderr, "%d: waited for a frame: %f ms\n", frame_no, (double)t / 1000000 );
			}
		}
		frame_no++;

#ifdef CAPTURE_NEED_SEQNO
		if(frame_seqno[buf_idx] > (frame_seqno[prev_buf_idx] + 1)) {
			fprintf( stderr, "frame(s) dropped. frame_no: %d, new: %u(time: %lld), prev:%u(time:%lld)\n",
					frame_no,
					frame_seqno[buf_idx], frame_timestamp[buf_idx],
					frame_seqno[prev_buf_idx], frame_timestamp[prev_buf_idx]);
		}
		if(frame_flags[buf_idx]) {
			if(frame_flags[buf_idx] & CAPTURE_FRAME_FLAG_INTERLACED) {
				fprintf(stderr, "frame_no: %d (seqno: %d), interlace frame (%s)\n",
						frame_no,
						frame_seqno[buf_idx],
						frame_flags[buf_idx] & CAPTURE_FRAME_FLAG_FIELD_BOTTOM? "bottom":"top");
			}
			if(frame_flags[buf_idx] & CAPTURE_FRAME_FLAG_ERROR) {
				fprintf(stderr, "frame_no: %d (seqno: %d), error frame\n",
						frame_no,
						frame_seqno[buf_idx]);
			}
		}
#endif
		if(capture_is_property( context, CAPTURE_PROPERTY_FRAME_FLAGS )) {
			if((frame_flags[buf_idx] & CAPTURE_FRAME_FLAG_INTERLACED) &&
					(deinterlace_mode == CAPTURE_DEINTERLACE_NONE_MODE)) {
				interlaced = (frame_flags[buf_idx] & CAPTURE_FRAME_FLAG_FIELD_BOTTOM)?
						SCREEN_INTERLACED_BOTTOM_FIELD:
						SCREEN_INTERLACED_TOP_FIELD;
				rc = screen_set_buffer_property_iv(screen_buf[buf_idx], SCREEN_PROPERTY_INTERLACED, &interlaced);
				if (rc) {
						perror( "screen_set_buffer_property_iv(SCREEN_PROPERTY_INTERLACED)");
						goto fail;
				}
			}
		}
		int rect[4] = { spos[0], spos[1], ssize[0], ssize[1] };
		// update screen
		rc = screen_post_window( screen_win, screen_buf[buf_idx], 1, rect, 0 );
		if (rc) {
			perror("screen_post_window");
			goto fail;
		}


		// Make the buffer identified by the pre_buf_idx as available for capturing.
		if(prev_buf_idx != -1) {
			capture_release_frame( context, prev_buf_idx );
		}
		prev_buf_idx = buf_idx;

		// update the color parameters every 100 frames
		if(color_test > 0 && !(frame_no % 10)) {
			rc = change_color_settings(context, color_test);
			if(rc == -1) {
				fprintf(stderr, "the specified color test is not supported!\n");
				color_test = 0;
			}
		}

		if(video_info_frame_cnt > 0 && !(frame_no % video_info_frame_cnt)) {
			get_video_info(context);
		}

		if(frame_no == total_frame_cnt) {
			// tell the driver to stop capturing (when capture_update() is called).
			capture_set_property_i( context, CAPTURE_ENABLE, 0 );

			// commit changes to the H/W
			if(capture_update( context, 0 ) == -1) {
				fprintf(stderr, "capture_update failed\n");
				goto fail;
			}

			if(reenable_delay_msec == -1) {
				char buffer[10];
				fprintf(stderr, "press enter to continue capturing, or type \"n\" to exit\n");
				fgets(buffer, sizeof(buffer), stdin);
				if(!strcmp(buffer, "n\n")) {
					goto end;
				}
			}
			else {
				time_t delay_t;
				time(&delay_t);
				fprintf( stderr, "delay at %s", ctime(&delay_t) );
				delay(reenable_delay_msec);
			}
			frame_no = 0;
			capture_set_property_i( context, CAPTURE_ENABLE, 1 );

			// commit changes to the H/W -- and start capturing...
			if(capture_update( context, 0 ) == -1) {
				fprintf(stderr, "capture_update failed\n");
				goto fail;
			}
		}
	}
end:
	rval = EXIT_SUCCESS;
fail:
	if(context != NULL) {
		capture_destroy_context( context);
	}
	if(screen_win != NULL) {
		screen_destroy_window(screen_win);
	}
	if(screen_ctx != NULL) {
		screen_destroy_context(screen_ctx);
	}
#ifdef HD_DEC_CLIENT
	if(use_resmgr) {
		hd_dec_client_fini();
	}
#endif
	return rval;
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn/product/branches/6.6.0/trunk/apps/screen/demos/vcapture-test/vcapture-test.c $ $Rev: 760287 $")
#endif
