/*
* Copyright (c) 2013 QNX Software Systems Limited.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <errno.h>
#include <fcntl.h>
#include <img/img.h>
#include <math.h>
#include <screen/screen.h>
#include <string.h>
#include <stdlib.h>

#include "input/screen_helpers.h"
#include "gestures/types.h"
#include "gestures/set.h"
#include "gestures/swipe.h"
#include "gestures/pinch.h"
#include "gestures/press_and_tap.h"
#include "gestures/rotate.h"
#include "gestures/two_finger_pan.h"
#include "gestures/tap.h"
#include "gestures/double_tap.h"
#include "gestures/triple_tap.h"
#include "gestures/long_press.h"
#include "gestures/two_finger_tap.h"

#define MIN_VIEWPORT_SIZE 128
#define MAX_VIEWPORT_SIZE 4096
screen_context_t screen_ctx;
screen_window_t screen_win;

struct gestures_set * set;
const char* img_path = "/usr/share/images/wallpaper.jpg"; /* path to image asset */
int viewport_pos[2] = { 0, 0 };
int viewport_size[2] = { 0, 0 };

/* The callback invoked when a gesture is recognized or updated. */
void
gesture_callback(gesture_base_t* gesture, mtouch_event_t* event, void* param, int async)
{
	/* Check if callback is invoked from Timer thread... */
    if (async) {
        fprintf(stderr,"Callback invoked from Timer thread.");
    }

    /* Perform actions based on gesture type */
    switch (gesture->type) {
        /* For a two-finger pan, the application copies information from the incoming gesture to a local gesture_tfpan_t structure 
         * and uses the information to adjust the position of the viewport on the display. 
         * In this case, the viewport is adjusted to follow the direction of the pan by determining the distance of the swipe 
         * and adding it to the current x and y position of the viewport */ 
        case GESTURE_TWO_FINGER_PAN: {
        	fprintf(stderr, "GESTURE_TWO_FINGER_PAN detected:  ");
            gesture_tfpan_t* tfpan = (gesture_tfpan_t*)gesture;
            fprintf(stderr,"Two Finger Pan: %d, %d", (tfpan->last_centroid.x - tfpan->centroid.x), (tfpan->last_centroid.y - tfpan->centroid.y));
            if (tfpan->last_centroid.x && tfpan->last_centroid.y) {
                viewport_pos[0] += (tfpan->last_centroid.x - tfpan->centroid.x) >> 1;
                viewport_pos[1] += (tfpan->last_centroid.y - tfpan->centroid.y) >> 1;
            }
            break;
        }
        /* For a pinch, in the gestures callback function, the application copies the gesture information into a local gesture_pinch_t structure 
         * and uses it to adjust the size of the viewport. This is done by first determining the change in distance between the two fingers 
         * at the start of the pinch and at the end of the pinch: */
        case GESTURE_PINCH: {
        	fprintf(stderr, "GESTURE_PINCH detected:  ");
            gesture_pinch_t* pinch = (gesture_pinch_t*)gesture;
            fprintf(stderr,"Pinch %d, %d", (pinch->last_distance.x - pinch->distance.x), (pinch->last_distance.y - pinch->distance.y));

            int dist_x = pinch->distance.x;
            int dist_y = pinch->distance.y;
            int last_dist_x = pinch->last_distance.x;
            int last_dist_y = pinch->last_distance.y;

             /* The application calculates the relative distance between the locations to determine the net increase or decrease of the viewport's size. 
              * If either of the relative distance values are greater than zero, the application adjusts the viewport size */
            int reldist = sqrt((dist_x)*(dist_x) + (dist_y)*(dist_y));
            int last_reldist = sqrt((last_dist_x)*(last_dist_x) + (last_dist_y)*(last_dist_y));

            if (reldist && last_reldist) {
                viewport_size[0] += (last_reldist - reldist) >> 1;
                viewport_size[1] += (last_reldist - reldist) >> 1;

                /* Size restrictions */
                if (viewport_size[0] < MIN_VIEWPORT_SIZE) {
                    viewport_size[0] = MIN_VIEWPORT_SIZE;
                } else if (viewport_size[0] > MAX_VIEWPORT_SIZE) {
                    viewport_size[0] = MAX_VIEWPORT_SIZE;
                }
                if (viewport_size[1] < MIN_VIEWPORT_SIZE) {
                    viewport_size[1] = MIN_VIEWPORT_SIZE;
                } else if (viewport_size[1] > MAX_VIEWPORT_SIZE) {
                    viewport_size[1] = MAX_VIEWPORT_SIZE;
                }

                /* Zoom into center of image */
                if (viewport_size[0] > MIN_VIEWPORT_SIZE && viewport_size[1] > MIN_VIEWPORT_SIZE &&
                        viewport_size[0] < MAX_VIEWPORT_SIZE && viewport_size[1] < MAX_VIEWPORT_SIZE) {
                    viewport_pos[0] -= (last_reldist - reldist) >> 2;
                    viewport_pos[1] -= (last_reldist - reldist) >> 2;
                }
            }
            break;
        }
        /* For other gestures, the application simply prints out the coordinates of the touch location */
        case GESTURE_TAP: {
        	fprintf(stderr, "GESTURE_TAP detected:  ");
            gesture_tap_t* tap = (gesture_tap_t*)gesture;
            fprintf(stderr,"Tap x:%d y:%d",tap->touch_coords.x,tap->touch_coords.y);
            break;
        }
        case GESTURE_DOUBLE_TAP: {
        	fprintf(stderr, "GESTURE_DOUBLE_TAP detected:  ");
            gesture_double_tap_t* d_tap = (gesture_double_tap_t*)gesture;
            fprintf(stderr, "Double tap first_touch x:%d y:%d", d_tap->first_touch.x,
                                                       d_tap->first_touch.y);
            fprintf(stderr, "Double tap first_release x:%d y:%d", d_tap->first_release.x,
                                                         d_tap->first_release.y);
            fprintf(stderr, "Double tap second_touch x:%d y:%d", d_tap->second_touch.x,
                                                        d_tap->second_touch.y);
            fprintf(stderr, "Double tap second_release x:%d y:%d", d_tap->second_touch.x,
                                                          d_tap->second_release.y);
            break;
        }
        case GESTURE_ROTATE: {
        	fprintf(stderr, "GESTURE_ROTATE detected:  ");
            gesture_rotate_t* rotate = (gesture_rotate_t*)gesture;
            if (rotate->angle != rotate->last_angle) {
            	fprintf(stderr, "Rotate: %d degs", rotate->angle - rotate->last_angle);
            } else {
                return;
            }
            break;
        }
        case GESTURE_SWIPE: {
        	fprintf(stderr, "GESTURE_SWIPE detected:  ");
            gesture_swipe_t* swipe = (gesture_swipe_t*)gesture;
            if (swipe->direction & GESTURE_DIRECTION_UP) {
            	fprintf(stderr, "up %d  ", swipe->last_coords.y - swipe->coords.y);
            } else if (swipe->direction & GESTURE_DIRECTION_DOWN) {
            	fprintf(stderr, "down %d  ", swipe->coords.y - swipe->last_coords.y);
            } else if (swipe->direction & GESTURE_DIRECTION_LEFT) {
            	fprintf(stderr, "left %d  ", swipe->last_coords.x - swipe->coords.x);
            } else if (swipe->direction & GESTURE_DIRECTION_RIGHT) {
            	fprintf(stderr, "right %d", swipe->coords.x - swipe->last_coords.x);
            }
            break;
        }
        case GESTURE_TRIPLE_TAP: {
        	fprintf(stderr, "GESTURE_TRIPLE_TAP detected:  ");
            gesture_triple_tap_t* t_tap = (gesture_triple_tap_t*)gesture;
            fprintf(stderr, "triple tap first_touch x:%d y:%d", t_tap->first_touch.x,
                                                       t_tap->first_touch.y);
            fprintf(stderr, "triple tap first_release x:%d y:%d", t_tap->first_release.x,
                                                         t_tap->first_release.y);
            fprintf(stderr, "triple tap second_touch x:%d y:%d", t_tap->second_touch.x,
                                                        t_tap->second_touch.y);
            fprintf(stderr, "triple tap second_release x:%d y:%d", t_tap->second_touch.x,
                                                          t_tap->second_release.y);
            fprintf(stderr, "triple tap third_touch x:%d y:%d", t_tap->third_touch.x,
                                                       t_tap->second_touch.y);
            fprintf(stderr, "triple tap third_release x:%d y:%d", t_tap->third_touch.x,
                                                         t_tap->second_release.y);
            break;
        }
        case GESTURE_PRESS_AND_TAP: {
        	fprintf(stderr, "GESTURE_PRESS_AND_TAP detected:  ");
            gesture_pt_t* pt_t = (gesture_pt_t*)gesture;
            fprintf(stderr, "Initial Press x:%d y:%d", pt_t->initial_coords[0].x,
                                              pt_t->initial_coords[0].y);
            fprintf(stderr, "Initial Tap x:%d y:%d", pt_t->initial_coords[1].x,
                                            pt_t->initial_coords[1].y);
            fprintf(stderr, "Press x:%d y:%d", pt_t->coords[0].x, pt_t->coords[0].y);
            fprintf(stderr, "Tap x:%d y:%d", pt_t->coords[1].x, pt_t->coords[1].y);
            break;
        }
        case GESTURE_TWO_FINGER_TAP: {
        	fprintf(stderr, "GESTURE_TWO_FINGER_TAP detected:  ");
            gesture_tft_t* tft_t = (gesture_tft_t*)gesture;
            fprintf(stderr, "Coordinates of touch event (finger 1) x:%d y:%d",
                                                          tft_t->touch_coords[0].x,
                                                          tft_t->touch_coords[0].y);
            fprintf(stderr, "Coordinates of touch event (finger 2) x:%d y:%d",
                                                          tft_t->touch_coords[1].x,
                                                          tft_t->touch_coords[1].y);
            fprintf(stderr, "Coordinates of release event (finger 1) x:%d y:%d",
                                                          tft_t->release_coords[0].x,
                                                          tft_t->release_coords[0].y);
            fprintf(stderr, "Coordinates of release event (finger 2) x:%d y:%d",
                                                          tft_t->release_coords[1].x,
                                                          tft_t->release_coords[1].y);
            fprintf(stderr, "Midpoint between two touches x:%d y:%d", tft_t->centroid.x,
                                                              tft_t->centroid.y);
            break;
        }
        case GESTURE_LONG_PRESS: {
        	fprintf(stderr, "GESTURE_LONG_PRESS detected:  ");
            gesture_long_press_t* lp_t = (gesture_long_press_t*)gesture;
            fprintf(stderr, "Long Press x:%d y:%d",lp_t->coords.x, lp_t->coords.y);
            printf("Timer ID:%d",lp_t->success_timer);
            break;
        }
        case GESTURE_USER: {
        	fprintf(stderr, "GESTURE_USER detected.");
            break;
        }
        default: {
            fprintf(stderr, "Unknown Gesture");
            break;
        }
    }
    fprintf(stderr,"\n");
}

/* Initialize the gestures set */
static void
init_gestures()
{
    /* The application must register the callback function with the gestures library for it to be invoked when a gesture occurs. 
     * This function allocates a gesture set for the application, using the set structure */
    set = gestures_set_alloc();
    if (NULL != set) {
        tap_gesture_alloc(NULL, gesture_callback, set);
        double_tap_gesture_alloc(NULL, gesture_callback, set);
        long_press_gesture_alloc(NULL, gesture_callback, set); 
        tfpan_gesture_alloc(NULL, gesture_callback, set); 
        pinch_gesture_alloc(NULL, gesture_callback, set); 
        triple_tap_gesture_alloc(NULL, gesture_callback, set); 
        swipe_gesture_alloc(NULL, gesture_callback, set); 
        rotate_gesture_alloc(NULL, gesture_callback, set); 
        tft_gesture_alloc(NULL, gesture_callback, set); 
        pt_gesture_alloc(NULL, gesture_callback, set); 

    } else {
        fprintf(stderr, "Failed to allocate gestures set\n");
    }
}

static void
gestures_cleanup()
{
    if (NULL != set) {
        gestures_set_free(set);
        set = NULL;
    }
}

static int decode_setup(uintptr_t data, img_t *img, unsigned flags)
{
    screen_window_t screen_win = (screen_window_t)data;
    screen_buffer_t screen_buf;
    int size[2];

    size[0] = img->w;
    size[1] = img->h;
    screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_BUFFER_SIZE, size);
    screen_create_window_buffers(screen_win, 1);

    screen_get_window_property_pv(screen_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)&screen_buf);
    screen_get_buffer_property_pv(screen_buf, SCREEN_PROPERTY_POINTER, (void **)&img->access.direct.data);
    screen_get_buffer_property_iv(screen_buf, SCREEN_PROPERTY_STRIDE, (int *)&img->access.direct.stride);

    img->flags |= IMG_DIRECT;
    return IMG_ERR_OK;
}

static void decode_abort(uintptr_t data, img_t *img)
{
    screen_window_t screen_win = (screen_window_t)data;
    screen_destroy_window_buffers(screen_win);
}

int
load_image(screen_window_t screen_win, const char *path)
{
    img_decode_callouts_t callouts;
    img_lib_t ilib = NULL;
    img_t img;
    int rc;

    rc = img_lib_attach(&ilib);
    if (rc != IMG_ERR_OK) {
		fprintf(stderr, "img_lib_attach() failed\n");
        return -1;
    }
    memset(&img, 0, sizeof(img));
    img.flags |= IMG_FORMAT;
    img.format = IMG_FMT_PKLE_XRGB8888;

    memset(&callouts, 0, sizeof(callouts));
    callouts.setup_f = decode_setup;
    callouts.abort_f = decode_abort;
    callouts.data = (uintptr_t)screen_win;

    rc = img_load_file(ilib, path, &callouts, &img);
    img_lib_detach(ilib);
    return rc == IMG_ERR_OK ? 0 : -1;
}


int
main(int argc, char **argv)
{
	/**
	 ** This is the size for an invisible exit button. We choose a value that's
	 ** big enough to be useable with touchscreens and pointer devices.
	 **/
	const int exit_area_size = 20;
	int val; /* used for simple property queries */
	int pos[2] = { 0, 0 }; /* position of the window on screen */
	int size[2] = { 0, 0 }; /* size of the window on screen */

    const int usage = SCREEN_USAGE_WRITE;

    screen_buffer_t screen_buf = NULL;
    int rect[4] = { 0, 0, 0, 0 };
    int screen_val = 0;
    int rc, i;
    screen_event_t screen_ev;
    mtouch_event_t mtouch_event;

    for (i = 1; i < argc; i++)
    {
        if (strncmp(argv[i], "-image=", strlen("-image=")) == 0)
            img_path = argv[i] + strlen("-image=");
    }

    /* Setup Screen context and the window for application */
    rc = screen_create_context(&screen_ctx, 0);
    if (rc) {
    	goto fail_create_context;
    }

    rc = screen_create_window(&screen_win, screen_ctx);
    if (rc) {
    	goto fail_create_window;
    }

    screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_USAGE, &usage);

    rc = load_image(screen_win, img_path);
    if (rc) {
        fprintf(stderr, "Error: failed to load image at %s\n", img_path);
    	goto fail_load_image;
    }

    screen_get_window_property_pv(screen_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)&screen_buf);
    screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_BUFFER_SIZE, rect+2);
    viewport_size[0] = rect[2];
    viewport_size[1] = rect[3];
    screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SOURCE_SIZE , viewport_size);

    screen_post_window(screen_win, screen_buf, 1, rect, 0);

    init_gestures();

	if (size[0] == -1 || size[1] == -1) {
		rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_BUFFER_SIZE, size);
		if (rc) {
			perror("screen_get_window_property_iv(SCREEN_PROPERTY_BUFFER_SIZE)");
			goto fail_create_event;
		}
	}

	rc = screen_create_event(&screen_ev);
    if (rc) {
        goto fail_create_event;
    }

    /* Main application loop to handle Screen events. */
    while (1) {
    	while (screen_get_event(screen_ctx, screen_ev, ~0L) == EOK) {
            rc = screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_TYPE, &screen_val);
            if (rc || screen_val == SCREEN_EVENT_NONE) {
            	fprintf(stderr, "Error: failed to get mtouch event\n");
                break;
            }
            switch (screen_val) {
                case SCREEN_EVENT_MTOUCH_TOUCH:
                case SCREEN_EVENT_MTOUCH_MOVE:
                case SCREEN_EVENT_MTOUCH_RELEASE: {
                    /* Note that this helper function is not under Screen library, but under Input library  */
                    rc = screen_get_mtouch_event(screen_ev, &mtouch_event, 0);
                    if (rc) {
                        fprintf(stderr, "Error: failed to get mtouch event\n");
                        continue;
                    }
                     /* The application must call gestures_set_process_event() when a touch event is detected and needs to be handled. */
                    rc = gestures_set_process_event(set, &mtouch_event, NULL);
                    break;
                }
				case SCREEN_EVENT_POINTER: {
					rc = screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_BUTTONS, &val);
					if (val) {
						rc = screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_POSITION, pos);
						if (rc){
						  rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_SIZE, size);
						  if (rc){
						     fprintf(stderr, "window width: %d, window height: %d\n", size[0], size[1]);
						     fprintf(stderr, "pointer x: %d, pointer y: %d\n", pos[0], pos[1]);
						     if (pos[0] >= size[0] - exit_area_size &&
							    pos[1] < exit_area_size) {
						    	fprintf(stderr, "Exiting\n");
							    goto end;
						     }
						  }
						}
					}
					break;
                }
                case SCREEN_EVENT_CLOSE:{
                	 fprintf(stderr, "Exiting.\n");
                	goto end;
                }
                default:
                	break;
            }

            /* Re-draw the screen after a screen event */
            screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SOURCE_POSITION , viewport_pos);
            screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SOURCE_SIZE , viewport_size);
            screen_flush_context(screen_ctx,0);
    	}
    }

    /* Clean up */
    end:
        rc = EXIT_SUCCESS;
    	screen_destroy_event(screen_ev);

    fail_create_event:
    	gestures_cleanup();

    fail_load_image:
    	screen_destroy_window(screen_win);
    fail_create_window:
    	screen_destroy_context(screen_ctx);
    fail_create_context:
    	return rc;
}
