/**
 ** sw-vsync
 ** Windowed vsync that uses software rendering.
 **
 ** Features:
 **  - configurable window size through the -size=[width]x[height] option
 **  - configurable window position through the -pos=[x],[y] option
 **  - adjustable swap interval through the -interval=[n] option;
 **    a swap interval of 0 lets the app run as fast as possible
 **    numbers of 1 or more limit the rate to the number of vsync periods
 **  - updates can be paused with the 'F' key
 **  - application exits if power or mode event invalidates window surface(s)
 **
 ** Please go to the following URL for updates to this tutorial:
 **   http://community.qnx.com/sf/scm/do/listRepositories/projects.graphics/scm
 **
 ** Copyright 2010, QNX Software Systems Ltd. All Rights Reserved
 **
 ** Permission to use, copy, modify, and distribute this software and its
 ** documentation for any purpose and without fee is hereby granted,
 ** provided that the above copyright notice appear in all copies and that
 ** both that copyright notice and this permission notice appear in
 ** supporting documentation.
 **
 ** This file is provided AS IS with no warranties of any kind.  The author
 ** shall have no liability with respect to the infringement of copyrights,
 ** trade secrets or any patents by this file or any part thereof.  In no
 ** event will the author be liable for any lost revenue or profits or
 ** other special, indirect and consequential damages.
 */

#define NAME "sw-vsync"

/**
 ** Include the header files for libraries we are using.
 **/

#include <ctype.h>          /* Header file for isdigit */
#include <stdio.h>          /* Header file for fprintf */
#include <stdlib.h>         /* Header file for EXIT_FAILURE, EXIT_SUCCESS, atoi */
#include <string.h>         /* Header file for strncmp */
#include <sys/keycodes.h>   /* Header file for KEYCODE_ESCAPE */
#include <time.h>           /* Header file for clock_gettime, timespec2nsec */
#include <screen/screen.h>  /* Header file for all native API (iow) calls */

/**
 ** This header file defines the data for an hourglass bitmap. This bitmap is
 ** used to draw the hourglass instead of using filled polygon because the
 ** image-based technique yields higher performance.
 **/

#include "hourglass.h"

/**
 ** We won't support more than two window buffers. This enables double-
 ** buffering. Triple buffering is not usefull in our case since we are
 ** using a software rasterizer instead of a GPU.
 **/

#define MAX_BUFFER_COUNT (2)

/**
 ** Make the sliding vertical blue bar 32 pixels wide. Any number would be
 ** fine here, as long as it is no larger than the width of the window.
 **/

const int barwidth = 32;

/**
 ** The draw functions doe the rendering. They refresh all the dirty pixels
 ** using simple software rendering. The width, height, and stride are provided
 ** as arguments to allow the function to give correct results for most window
 ** sizes. The red, green, blue, and alpha arguments represent the number of
 ** bit shifts to the left necessary to get to the first bit of the red, green,
 ** blue, and alpha components respectively. The rendering is done in three
 ** parts. The first part clears the screen. This is only done once when the
 ** first frame is rendered. The second part animates the vertical blue bar.
 ** The last part draws the hourglass shape.
 **/

static void draw_rgba5551(int16_t *buffer,  /* a pointer to the pbuffer */
						  int width,        /* width of the pbuffer */
						  int height,       /* height of the pbuffer */
						  int stride,       /* number of bytes between rows */
						  int red,          /* shift offset for red component */
						  int green,        /* shift offset for green component */
						  int blue,         /* shift offset for blue component */
						  int alpha,        /* shift offset for alpha component */
						  int *rects,	    /* (x, y, width, height) of 2 dirty rects */
						  unsigned char redval,
						  unsigned char greenval,
						  unsigned char blueval,
						  unsigned char alphaval)
{
	static int frame = 0;   /* frame counter used for the position of the bar */
	int16_t bg;             /* background yellow in packed format */
	int16_t bar;            /* blue color in packed format */
	int16_t grey;           /* gray color in packed format */
	int tmp;                /* use to calculate the position of the bar */
	int16_t *p;             /* points to a pixel in the buffer */
	int i;                  /* line counter */
	int j;                  /* pixel counter */

	/* use stride expressed in pixels */
	stride >>= 1;

	/* The background colour is changeable */
	bg = ((redval & 0x1f) << red) | ((greenval & 0x1f) << green) | ((blueval & 0x1f) << blue) | ((alphaval & 0x01) << alpha);

	/* The bar is blue: alpha=ff, red=0, green=0, blue=ff */
	bar = (0x1f << blue) | (0x1 << alpha);

	/* The hourglass is gray; alpha=ff, red=a0, green=a0, blue=a0 */
	grey = (0x14 << red) | (0x14 << green) | (0x14 << blue) | (0x1 << alpha);

	/**
	 ** Because the buffer contents are preserved by eglSwapBuffers and during
	 ** the mapping process, we only have to clear the window once.
	 **/

	if (frame == 0) {
		/* Loop through all lines in the buffer */
		for (i = 0, p = buffer; i < height; i++, p += stride - width) {
			/* Skip the pixels that will be covered by the vertical bar */
			p += barwidth;

			/* Do all the remaining pixels in a line */
			for (j = barwidth; j < width; j++, p++) {
				/* Write yellow to the buffer */
				*p = bg;
			}
		}
	}

	/**
	 ** We use tmp to make sure the vertical bar wraps when it touches the
	 ** right edge of the window instead of moving into oblivion.
	 **/
	tmp = frame % (width - barwidth);

	/**
	 ** When the bar wraps around, tmp will be 0 except for the first frame.
	 ** In this case, we must clear the previous bar location with yellow and
	 ** redraw and complete new bar at the left edge of the window. When we
	 ** don't wrap around, we can simply erase the column on the left side of
	 ** the bar with yellow and draw the column on the right side of the bar
	 ** blue.
	 **/

	if (tmp != 0) {
		/* Lopp through all lines in the buffer */
		for (i = 0; i < height; i++) {
			/* Clear the first pixel with yellow */
			buffer[i * stride + tmp - 1] = bg;

			/* Add blue to last pixel */
			buffer[i * stride + tmp + barwidth - 1] = bar;
		}
	} else {
		/**
		 ** When we do the first frame, there is no vertical bar on the right
		 ** side of the window to erase.
		 **/

		if (frame != 0) {
			/* clear the entire previous bar location with yellow */
			for (i = 0; i < height; i++) {
				p = buffer + i * stride + width - barwidth;
				for (j = 0; j < barwidth; j++, p++) {
					*p = bg;
				}
			}
		}

		/* draw the new bar */
		for (i = 0; i < height; i++) {
			p = buffer + i * stride;
			for (j = 0; j < barwidth; j++, p++) {
				*p = bar;
			}
		}
	}

	/**
	 ** We only redraw the hourglass if the vertical bar partially intersects
	 ** it or when we draw the first frame. Because the barwidth is such that
	 ** on the first frame the bar intersects with the hourglass, we can only
	 ** test for one condition.
	 **/

	if (tmp + barwidth - 1 >= 10 && tmp < 110) {
		/* We only have to go through the lines where the hourglass is present */
		for (i = 0; i < hourglass_height; i++) {
			/* Move the pointer to the start of the hourglass */
			p = buffer + (i + 10) * stride + 10;
			for (j = 0; j < hourglass_width; j++, p++) {
				/* hourglass_data contains a space to indicate transparency */
				if (hourglass_data[i * hourglass_width + j] != ' ') {
					*p = grey;
				}
			}
		}
	}

	/* we only have to update the x-axis of the dirty rects */
	if (tmp != 0) {
		rects[0] = tmp - 1;
		rects[2] = 1;
		rects[4] = tmp + barwidth - 1;
		rects[6] = 1;
	} else {
		rects[0] = 0;
		rects[2] = barwidth;
		if (frame != 0) {
			rects[4] = width - barwidth - 1;
			rects[6] = barwidth;
		} else {
			rects[4] = barwidth;
			rects[6] = width - barwidth;
		}
	}

	frame++;
}

static void draw_rgb565(int16_t *buffer,  /* a pointer to the pbuffer */
						int width,        /* width of the pbuffer */
						int height,       /* height of the pbuffer */
						int stride,       /* number of bytes between rows */
						int red,          /* shift offset for red component */
						int green,        /* shift offset for green component */
						int blue,         /* shift offset for blue component */
						int *rects,
						unsigned char redval,
						unsigned char greenval,
						unsigned char blueval)       /* (x, y, width, height) of 2 dirty rects */
{
	static int frame = 0;   /* frame counter used for the position of the bar */
	int16_t bg;             /* background yellow in packed format */
	int16_t bar;            /* blue color in packed format */
	int16_t grey;           /* gray color in packed format */
	int tmp;                /* use to calculate the position of the bar */
	int16_t *p;             /* points to a pixel in the buffer */
	int i;                  /* line counter */
	int j;                  /* pixel counter */

	/* use stride expressed in pixels */
	stride >>= 1;

	/* The background is colour is changeable */
	bg = ((redval & 0x1f) << red) | ((greenval & 0x3f) << green) | ((blueval & 0x1f) << blue);

	/* The bar is blue: alpha=ff, red=0, green=0, blue=ff */
	bar = (0x1f << blue);

	/* The hourglass is gray; alpha=ff, red=a0, green=a0, blue=a0 */
	grey = (0x14 << red) | (0x28 << green) | (0x14 << blue);

	/**
	 ** Because the buffer contents are preserved by eglSwapBuffers and during
	 ** the mapping process, we only have to clear the window once.
	 **/

	if (frame <= 0) {
		/* Loop through all lines in the buffer */
		for (i = 0, p = buffer; i < height; i++, p += stride - width) {
			/* Skip the pixels that will be covered by the vertical bar */
			p += barwidth;

			/* Do all the remaining pixels in a line */
			for (j = barwidth; j < width; j++, p++) {
				/* Write yellow to the buffer */
				*p = bg;
			}
		}
	}

	/**
	 ** We use tmp to make sure the vertical bar wraps when it touches the
	 ** right edge of the window instead of moving into oblivion.
	 **/
	tmp = frame % (width - barwidth);

	/**
	 ** When the bar wraps around, tmp will be 0 except for the first frame.
	 ** In this case, we must clear the previous bar location with yellow and
	 ** redraw and complete new bar at the left edge of the window. When we
	 ** don't wrap around, we can simply erase the column on the left side of
	 ** the bar with yellow and draw the column on the right side of the bar
	 ** blue.
	 **/

	if (tmp != 0) {
		/* Lopp through all lines in the buffer */
		for (i = 0; i < height; i++) {
			/* Clear the first pixel with yellow */
			buffer[i * stride + tmp - 1] = bg;

			/* Add blue to last pixel */
			buffer[i * stride + tmp + barwidth - 1] = bar;
		}
	} else {
		/**
		 ** When we do the first frame, there is no vertical bar on the right
		 ** side of the window to erase.
		 **/

		if (frame != 0) {
			/* clear the entire previous bar location with yellow */
			for (i = 0; i < height; i++) {
				p = buffer + i * stride + width - barwidth - 1;
				for (j = 0; j < barwidth; j++, p++) {
					*p = bg;
				}
			}
		}

		/* draw the new bar */
		for (i = 0; i < height; i++) {
			p = buffer + i * stride;
			for (j = 0; j < barwidth; j++, p++) {
				*p = bar;
			}
		}
	}

	/**
	 ** We only redraw the hourglass if the vertical bar partially intersects
	 ** it or when we draw the first frame. Because the barwidth is such that
	 ** on the first frame the bar intersects with the hourglass, we can only
	 ** test for one condition.
	 **/

	if (tmp + barwidth - 1 >= 10 && tmp < 110) {
		/* We only have to go through the lines where the hourglass is present */
		for (i = 0; i < hourglass_height; i++) {
			/* Move the pointer to the start of the hourglass */
			p = buffer + (i + 10) * stride + 10;
			for (j = 0; j < hourglass_width; j++, p++) {
				/* hourglass_data contains a space to indicate transparency */
				if (hourglass_data[i * hourglass_width + j] != ' ') {
					*p = grey;
				}
			}
		}
	}

	/* we only have to update the x-axis of the dirty rects */
	if (tmp != 0) {
		rects[0] = tmp - 1;
		rects[2] = 1;
		rects[4] = tmp + barwidth - 1;
		rects[6] = 1;
	} else {
		rects[0] = 0;
		rects[2] = barwidth;
		if (frame != 0) {
			rects[4] = width - barwidth - 1;
			rects[6] = barwidth;
		} else {
			rects[4] = barwidth;
			rects[6] = width - barwidth;
		}
	}

	frame++;
}

static void draw_rgba8888(int32_t *buffer,  /* a pointer to the pbuffer */
						  int width,        /* width of the pbuffer */
						  int height,       /* height of the pbuffer */
						  int stride,       /* number of bytes between rows */
						  int red,          /* shift offset for red component */
						  int green,        /* shift offset for green component */
						  int blue,         /* shift offset for blue component */
						  int alpha,        /* shift offset for alpha component */
						  int *rects,       /* (x, y, width, height) of 2 dirty rects */
						  unsigned char redval,
						  unsigned char greenval,
						  unsigned char blueval,
						  unsigned char alphaval)
{
	static int frame = 0;   /* frame counter used for the position of the bar */
	int32_t bg;             /* background yellow in packed format */
	int32_t bar;            /* blue color in packed format */
	int32_t grey;           /* gray color in packed format */
	int tmp;                /* use to calculate the position of the bar */
	int32_t *p;             /* points to a pixel in the buffer */
	int i;                  /* line counter */
	int j;                  /* pixel counter */

	/* use stride expressed in pixels */
	stride >>= 2;

	/* background has changeable rbga values */
	bg = (redval << red) | (greenval << green) | (blueval << blue) | (alphaval << alpha);

	/* The bar is blue: alpha=ff, red=0, green=0, blue=ff */
	bar = (0xff << blue) | (0xff << alpha);

	/* The hourglass is gray; alpha=ff, red=a0, green=a0, blue=a0 */
	grey = (0xa0 << red) | (0xa0 << green) | (0xa0 << blue) | (0xff << alpha);

	/**
	 ** Because the buffer contents are preserved by eglSwapBuffers and during
	 ** the mapping process, we only have to clear the window once.
	 **/

	if (frame == 0) {
		/* Loop through all lines in the buffer */
		for (i = 0, p = buffer; i < height; i++, p += stride - width) {
			/* Skip the pixels that will be covered by the vertical bar */
			p += barwidth;

			/* Do all the remaining pixels in a line */
			for (j = barwidth; j < width; j++, p++) {
				/* Write yellow to the buffer */
				*p = bg;
			}
		}
	}

	/**
	 ** We use tmp to make sure the vertical bar wraps when it touches the
	 ** right edge of the window instead of moving into oblivion.
	 **/
	tmp = frame % (width - barwidth);

	/**
	 ** When the bar wraps around, tmp will be 0 except for the first frame.
	 ** In this case, we must clear the previous bar location with yellow and
	 ** redraw and complete new bar at the left edge of the window. When we
	 ** don't wrap around, we can simply erase the column on the left side of
	 ** the bar with yellow and draw the column on the right side of the bar
	 ** blue.
	 **/

	if (tmp != 0) {
		/* Lopp through all lines in the buffer */
		for (i = 0; i < height; i++) {
			/* Clear the first pixel with yellow */
			buffer[i * stride + tmp - 1] = bg;

			/* Add blue to last pixel */
			buffer[i * stride + tmp + barwidth - 1] = bar;
		}
	} else {
		/**
		 ** When we do the first frame, there is no vertical bar on the right
		 ** side of the window to erase.
		 **/

		if (frame != 0) {
			/* clear the entire previous bar location with yellow */
			for (i = 0; i < height; i++) {
				p = buffer + i * stride + width - barwidth - 1;
				for (j = 0; j < barwidth; j++, p++) {
					*p = bg;
				}
			}
		}

		/* draw the new bar */
		for (i = 0; i < height; i++) {
			p = buffer + i * stride;
			for (j = 0; j < barwidth; j++, p++) {
				*p = bar;
			}
		}
	}

	/**
	 ** We only redraw the hourglass if the vertical bar partially intersects
	 ** it or when we draw the first frame. Because the barwidth is such that
	 ** on the first frame the bar intersects with the hourglass, we can only
	 ** test for one condition.
	 **/

	if (tmp + barwidth - 1 >= 10 && tmp < 110) {
		/* We only have to go through the lines where the hourglass is present */
		for (i = 0; i < hourglass_height; i++) {
			/* Move the pointer to the start of the hourglass */
			p = buffer + (i + 10) * stride + 10;
			for (j = 0; j < hourglass_width; j++, p++) {
				/* hourglass_data contains a space to indicate transparency */
				if (hourglass_data[i * hourglass_width + j] != ' ') {
					*p = grey;
				}
			}
		}
	}

	/* we only have to update the x-axis of the dirty rects */
	if (tmp != 0) {
		rects[0] = tmp - 1;
		rects[2] = 1;
		rects[4] = tmp + barwidth - 1;
		rects[6] = 1;
	} else {
		rects[0] = 0;
		rects[2] = barwidth;
		if (frame != 0) {
			rects[4] = width - barwidth - 1;
			rects[6] = barwidth;
		} else {
			rects[4] = barwidth;
			rects[6] = width - barwidth;
		}
	}

    frame++;
}

static void
draw_point( screen_context_t ctx, screen_buffer_t buf, int x, int y, unsigned long color )
{
	int fb[] = {
		SCREEN_BLIT_COLOR, color,
		SCREEN_BLIT_DESTINATION_HEIGHT, 1,
		SCREEN_BLIT_DESTINATION_WIDTH, 1,
		SCREEN_BLIT_DESTINATION_Y, y,
		SCREEN_BLIT_DESTINATION_X, x,
		SCREEN_BLIT_END
	};
	screen_fill( ctx, buf, fb );
}

static void
draw_hline_f( screen_context_t ctx, screen_buffer_t buf, int width, int y, unsigned long color )
{
	int fb[] = {
		SCREEN_BLIT_COLOR, color,
		SCREEN_BLIT_DESTINATION_HEIGHT, 1,
		SCREEN_BLIT_DESTINATION_WIDTH, width,
		SCREEN_BLIT_DESTINATION_Y, y,
		SCREEN_BLIT_DESTINATION_X, 0,
		SCREEN_BLIT_END
	};
	screen_fill( ctx, buf, fb );
}

static void
draw_vline_f( screen_context_t ctx, screen_buffer_t buf, int height, int x, unsigned long color )
{
	int fb[] = {
		SCREEN_BLIT_COLOR, color,
		SCREEN_BLIT_DESTINATION_HEIGHT, height,
		SCREEN_BLIT_DESTINATION_WIDTH, 1,
		SCREEN_BLIT_DESTINATION_Y, 0,
		SCREEN_BLIT_DESTINATION_X, x,
		SCREEN_BLIT_END
	};
	screen_fill( ctx, buf, fb );
}

static void
draw_grid( screen_context_t ctx, screen_buffer_t buf, int width, int height,
	int red_shift, int green_shift, int blue_shift, int alpha_shift
) {
	int i;
	unsigned long white = (0xff << red_shift) | (0xff << blue_shift) | (0xff << green_shift) | (0xFF << alpha_shift);
	unsigned long rd = (0xff << red_shift) | (0x00 << blue_shift) | (0x00 << green_shift) | (0xFF << alpha_shift);

	// lines
	for( i = 40; i < height; i += 40 ) {
		draw_hline_f( ctx, buf, width, i, white );
	}

	for( i = 40; i < width; i += 40 ) {
		draw_vline_f( ctx, buf, height, i, white );
	}

	// dots
	for( i = 0; i < 40; i++ ) {
		draw_point( ctx, buf, i, i, white );
	}
	
	for( i = 0; i < 40; i++ ) {
		draw_point( ctx, buf, width-i, height-i, white );
	}
	// bounding box
	draw_hline_f( ctx, buf, width, 0, rd );
	draw_hline_f( ctx, buf, width, height-1, rd );
	draw_vline_f( ctx, buf, height, 0, rd );
	draw_vline_f( ctx, buf, height, width-1, rd );
}
int main(int argc, char **argv)
{
	/**
	 ** This is the size for an invisible exit button. We choose a value that's
	 ** big enough to be useable with touchscreens and pointer devices.
	 **/

	const int exit_area_size = 20;

	/**
	 ** This structure contains the blit attributes that will be passed to
	 ** screen_blit calls. Instead of getting its information from function
	 ** arguments, the function gets them from a list of tokenized attributes.
	 ** This allows additional parameters to be added while remaining backwards
	 ** compatible with applications using older versions of the native API. We
	 ** will change these attributes often, so we will also use a structure
	 ** with named attribute/value pairs to avoid writing at the wrong index
	 ** location.
	 **/

	struct {
		int src_x[2];
		int src_y[2];
		int src_width[2];
		int src_height[2];
		int dst_x[2];
		int dst_y[2];
		int dst_width[2];
		int dst_height[2];
		int none;
	} screen_blit_attr = {
		.src_x = { SCREEN_BLIT_SOURCE_X, 0 },                /* Offset from the right edge of src */
		.src_y = { SCREEN_BLIT_SOURCE_Y, 0 },                /* Offset from the top edge of src */
		.src_width = { SCREEN_BLIT_SOURCE_WIDTH, 0 },        /* Number of columns to copy from src */
		.src_height = { SCREEN_BLIT_SOURCE_HEIGHT, 0 },      /* Number of rows top copy from src */
		.dst_x = { SCREEN_BLIT_DESTINATION_X, 0 },           /* Offset from the right edge in dst */
		.dst_y = { SCREEN_BLIT_DESTINATION_Y, 0 },           /* Offset from the top edge in dst */
		.dst_width = { SCREEN_BLIT_DESTINATION_WIDTH, 0 },   /* Number of columns to copy into dst */
		.dst_height = { SCREEN_BLIT_DESTINATION_HEIGHT, 0 }, /* Number of rows to copy into dst */
		.none = SCREEN_BLIT_END                              /* End of list */
	};

	screen_context_t screen_ctx;                   /* connection to screen windowing system */
	screen_display_t *screen_disp;                 /* native handle for our display */
	screen_window_t screen_win;                    /* native handle for our window */
	screen_pixmap_t screen_pix;                    /* native handle for off-screen rendering */
	screen_buffer_t screen_pbuf;                   /* pixmap's buffer */
	screen_buffer_t screen_wbuf[MAX_BUFFER_COUNT]; /* renderable buffers for the window */
	screen_event_t screen_ev;                      /* handle used to pop events from our queue */
	int size[2] = { -1, -1 };                      /* size of the window on screen */
	int pos[2] = { 0, 0 };                         /* position of the window on screen */
	int format = 0;                                /* pixel format for our window and pixmap */
	int interval = 1;                              /* swap interval */
	int nbuffers = 1;                              /* number of buffers (1 or 2) */
	int transp = SCREEN_TRANSPARENCY_NONE;         /* window transparency mode */
	//int ndisplays;                                 /* number of displays */
	int val;                                       /* used for simple property queries */
	const char *display = NULL;                    /* the display to create our window on */
	const char *tok;                               /* used to process command line arguments */
	int rval = EXIT_FAILURE;                       /* application exits with value stored here */
	int rc;                                        /* store return value from functions */
	int i;                                         /* loop/frame counter */
	int stride;                                    /* size of each pixmap line in bytes */
	void *pointer;                                 /* virtual address of the pixmap buffer */
	int pause = 0;                                 /* indicates if rendering should be paused */
	int rects[8];                                  /* stores up to 2 dirty rectangles */
	int pipeline = -1;
	int frame_limit_counter = -1;
	int zorder = 0;
	int opt_grid = 0;
	unsigned int bg_rgba[4] = {0xff, 0xff, 0x0, 0xff};   /* stores rgba values from argument */

	int screen_ndisplays;		/* number of displays */
	int screen_val;


	/**
	 ** We start by processing the command line arguments. The first argument
	 ** is skipped because it contains the name of the program. Arguments
	 ** follow the syntax -(option)=(value).
	 **/

	for (i = 1; i < argc; i++) {
		if (strncmp(argv[i], "-size=", strlen("-size=")) == 0) {
			/**
			 ** The syntax of the size option is -size=(width)x(height).
			 **/

			tok = argv[i] + strlen("-size=");
			size[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			size[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-pos=", strlen("-pos=")) == 0) {
			/**
			 ** The syntax of the pos option is -pos=(x),(y).
			 **/

			tok = argv[i] + strlen("-pos=");
			pos[0] = atoi(tok);
			while (*tok >= '0' && *tok <= '9') {
				tok++;
			}
			pos[1] = atoi(tok+1);
		} else if (strncmp(argv[i], "-format=", strlen("-format=")) == 0) {
			/**
			 ** The syntax of the format option is:
			 **   -format=rgb565, or
			 **   -format=rgba5551, or
			 **   -format=rgbx5551, or
			 **   -format=rgba8888, or
			 **   -format=rgbx8888.
			 **
			 ** These are the only formats supported by this tutorial.
			 **/

			tok = argv[i] + strlen("-format=");
			if (strcmp(tok, "rgb565") == 0) {
				format = SCREEN_FORMAT_RGB565;
			} else if (strcmp(tok, "rgba5551") == 0) {
				format = SCREEN_FORMAT_RGBA5551;
			} else if (strcmp(tok, "rgbx5551") == 0) {
				format = SCREEN_FORMAT_RGBX5551;
			} else if (strcmp(tok, "rgba8888") == 0) {
				format = SCREEN_FORMAT_RGBA8888;
			} else if (strcmp(tok, "rgbx8888") == 0) {
				format = SCREEN_FORMAT_RGBX8888;
			} else {
				fprintf(stderr, "invalid format value: %s\n", tok);
			}
		} else if (strncmp(argv[i], "-transparency=", strlen("-transparency")) == 0) {
			/**
			 ** The syntax for the transparency option is:
			 **   -transparency=none, or
			 **   -transparency=test, or
			 **   -transparency=src.
			 */

			tok = argv[i] + strlen("-transparency=");
			if (strcmp(tok, "none") == 0) {
				transp = SCREEN_TRANSPARENCY_NONE;
			} else if (strcmp(tok, "test") == 0) {
				transp = SCREEN_TRANSPARENCY_TEST;
			} else if (strcmp(tok, "src") == 0) {
				transp = SCREEN_TRANSPARENCY_SOURCE_OVER;
			} else {
				fprintf(stderr, "invalid transparency value: %s\n", tok);
			}
		} else if (strcmp(argv[i], "-single-buffer") == 0) {
			/**
			 ** The -single-buffer option on the command line will cause the
			 ** rendering to be done to a single on-screen buffer. There are
			 ** typically artifacts associated with single-buffered rendering
			 ** caused by rendering to a visible surface.
			 **/

			nbuffers = 1;
		} else if (strcmp(argv[i], "-double-buffer") == 0) {
			/**
			 ** The -double-buffer option on the command line will cause the
			 ** rendering to be on alternating back buffers. This eliminates
			 ** artifacts associated with single-buffered rendering.
			 **/

			nbuffers = 2;
		} else if (strncmp(argv[i], "-display=", strlen("-display=")) == 0) {
			/**
			 ** The syntax of the display option is -display=(number) or
			 ** -display=(string), where string can be one of the following
			 ** display port types:
			 **   internal,
			 **   composite,
			 **   svideo,
			 **   YPbPr,
			 **   rgb,
			 **   rgbhv,
			 **   dvi,
			 **   hdmi,
			 **   otehr.
			 **/

			display = argv[i] + strlen("-display=");
		} else if (strncmp(argv[i], "-interval=", strlen("-interval=")) == 0) {
			/**
			 ** The syntax of the interval option is -interval=(number). All
			 ** we need is to do is convert the number that starts after
			 ** the '='.
			 **/

			interval = atoi(argv[i] + strlen("-interval="));
		} else if (strncmp(argv[i], "-pipeline=", strlen("-pipeline=")) == 0) {
			pipeline = atoi(argv[i] + strlen("-pipeline="));
		} else if (strncmp(argv[i], "-frame-limit=", strlen("-frame-limit=")) == 0) {
			frame_limit_counter = atoi(argv[i] + strlen("-frame-limit="));
		} else if (strncmp(argv[i], "-zorder=", strlen("-zorder=")) == 0) {
			zorder = atoi(argv[i] + strlen("-zorder="));
		} else if( strncmp(argv[i], "-grid", strlen("-grid")) == 0 ) {
			opt_grid = 1;
		} else if( strncmp(argv[i], "-bg_rgba=", strlen("-bg_rgba=")) == 0 ) {
			/**
			 ** looking for rgba values for background window
			 ** -bg_rgba=(string), where string can be the following
			 ** red,green,blue,alpha
			 ** all integer values
			 **/

			tok = argv[i] + strlen("-bg_rgba=");

			if( 4 != sscanf( tok, "%i,%i,%i,%i", &bg_rgba[0], &bg_rgba[1], &bg_rgba[2], &bg_rgba[3] ) ) {
				fprintf( stderr,"-bg_rgba=%s is not correctly formatted\n", tok );
			}
		} else {
			/**
			 ** Make sure we say something instead of silently ignoring a
			 ** command line option.
			 **/

			fprintf(stderr, "invalid command line option: %s\n", argv[i]);
		}
	}

	rc = screen_create_context(&screen_ctx, 0);
	if (rc) {
		perror("screen_context_create");
		goto fail1;
	}

	rc = screen_create_window(&screen_win, screen_ctx);
	if (rc) {
		perror("screen_create_window");
		goto fail2;
	}

	int layout = 0;
	rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_SELF_LAYOUT, &layout);
	if(!rc) {
		printf("sw-vsync self layout %d\n", layout);
	}

	if (display) {
		rc = screen_get_context_property_iv(screen_ctx, SCREEN_PROPERTY_DISPLAY_COUNT, 
			&screen_ndisplays);
		if (rc) {
			perror("screen_get_context_property_iv(SCREEN_PROPERTY_DISPLAY_COUNT)");
			goto fail3;
		}

		screen_disp = calloc(screen_ndisplays, sizeof(*screen_disp));
		if (screen_disp == NULL) {
			fprintf(stderr, "could not allocate memory for display list\n");
			goto fail3;
		}

		rc = screen_get_context_property_pv(screen_ctx, SCREEN_PROPERTY_DISPLAYS, 
			(void **)screen_disp);
		if (rc) {
			perror("screen_get_context_property_ptr(SCREEN_PROPERTY_DISPLAYS)");
			free(screen_disp);
			goto fail3;
		}

		if (isdigit(*display)) {
			int want_id = atoi(display);
			for (i = 0; i < screen_ndisplays; ++i) {
				int actual_id = 0;  // invalid
				(void)screen_get_display_property_iv(screen_disp[i],
						SCREEN_PROPERTY_ID, &actual_id);
				if (want_id == actual_id) {
					break;
				}
			}
		} else {
			int type = -1;
			if (strcmp(display, "internal") == 0) {
				type = SCREEN_DISPLAY_TYPE_INTERNAL;
			} else if (strcmp(display, "composite") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPOSITE;
			} else if (strcmp(display, "svideo") == 0) {
				type = SCREEN_DISPLAY_TYPE_SVIDEO;
			} else if (strcmp(display, "YPbPr") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_YPbPr;
			} else if (strcmp(display, "rgb") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_RGB;
			} else if (strcmp(display, "rgbhv") == 0) {
				type = SCREEN_DISPLAY_TYPE_COMPONENT_RGBHV;
			} else if (strcmp(display, "dvi") == 0) {
				type = SCREEN_DISPLAY_TYPE_DVI;
			} else if (strcmp(display, "hdmi") == 0) {
				type = SCREEN_DISPLAY_TYPE_HDMI;
			} else if (strcmp(display, "other") == 0) {
				type = SCREEN_DISPLAY_TYPE_OTHER;
			} else {
				fprintf(stderr, "unknown display type %s\n", display);
				free(screen_disp);
				goto fail3;
			}
			for (i = 0; i < screen_ndisplays; i++) {
				screen_get_display_property_iv(screen_disp[i], 
					SCREEN_PROPERTY_TYPE, &screen_val);
				if (screen_val == type) {
					break;
				}
			}
		}
		if (i >= screen_ndisplays) {
			fprintf(stderr, "couldn't find display %s\n", display);
			free(screen_disp);
			goto fail3;
		}

		rc = screen_set_window_property_pv(screen_win, SCREEN_PROPERTY_DISPLAY, 
		(void **)&screen_disp[i]);
			if (rc) {
			perror("screen_set_window_property_ptr(SCREEN_PROPERTY_DISPLAY)");
			free(screen_disp);
			goto fail3;
		}

		free(screen_disp);
	}

	val = SCREEN_USAGE_NATIVE;
	if (pipeline != -1) {
		val |= SCREEN_USAGE_OVERLAY;
	}
	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_USAGE, &val);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_USAGE)");
		goto fail3;
	}

	if (format) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_FORMAT, &format);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_FORMAT)");
			goto fail3;
		}
	}

	if (size[0] != -1 && size[1] != -1) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SIZE, size);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_SIZE)");
			goto fail3;
		}
	}

	if (pos[0] != 0 || pos[1] != 0) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_POSITION, pos);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_POSITION)");
			goto fail3;
		} else {
			printf("setting position %d %d %d\n",SCREEN_PROPERTY_POSITION,pos[0],pos[1]);
		}
	}

	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_TRANSPARENCY, &transp);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_TRANSPARENCY)");
		goto fail3;
	}
	
	rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_ZORDER, &zorder);
	if (rc) {
		perror("screen_set_window_property_iv(SCREEN_PROPERTY_ZORDER)");
		goto fail3;
	}

	// Set ID string for debugging via /dev/screen.
	{
		const char *idstr = NAME;
		screen_set_window_property_cv(screen_win,
				SCREEN_PROPERTY_ID_STRING, strlen(idstr), idstr);
	}

	rc = screen_create_window_buffers(screen_win, nbuffers);
	if (rc) {
		perror("screen_create_window_buffers");
		goto fail3;
	}

	rc = screen_create_pixmap(&screen_pix, screen_ctx);
	if (rc) {
		perror("screen_create_pixmap");
		goto fail3;
	}

	val = SCREEN_USAGE_WRITE | SCREEN_USAGE_NATIVE;
	rc = screen_set_pixmap_property_iv(screen_pix, SCREEN_PROPERTY_USAGE, &val);
	if (rc) {
		perror("screen_set_pixmap_property_iv(SCREEN_PROPERTY_USAGE)");
		goto fail3;
	}

	if (!format) {
		rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_FORMAT, &format);
		if (rc) {
			perror("screen_get_window_property_iv(SCREEN_PROPERTY_FORMAT)");
			goto fail4;
		}
	}

	rc = screen_set_pixmap_property_iv(screen_pix, SCREEN_PROPERTY_FORMAT, &format);
	if (rc) {
		perror("screen_set_pixmap_property_iv(SCREEN_PROPERTY_FORMAT)");
		goto fail4;
	}

	if (size[0] == -1 || size[1] == -1) {
		rc = screen_get_window_property_iv(screen_win, SCREEN_PROPERTY_BUFFER_SIZE, size);
		if (rc) {
			perror("screen_get_window_property_iv(SCREEN_PROPERTY_BUFFER_SIZE)");
			goto fail4;
		}
	}

	rc = screen_set_pixmap_property_iv(screen_pix, SCREEN_PROPERTY_BUFFER_SIZE, size);
	if (rc) {
		perror("screen_set_pixmap_property_iv(SCREEN_PROPERTY_SIZE)");
		goto fail4;
	}

	// Set ID string for debugging via /dev/screen.
	{
		const char *idstr = NAME;
		screen_set_pixmap_property_cv(screen_pix,
				SCREEN_PROPERTY_ID_STRING, strlen(idstr), idstr);
	}

	rc = screen_create_pixmap_buffer(screen_pix);
	if (rc) {
		perror("screen_create_pixmap_buffer");
		goto fail4;
	}

	rc = screen_get_pixmap_property_pv(screen_pix, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)&screen_pbuf);
	if (rc) {
		perror("screen_get_pixmap_property_pv(SCREEN_PROPERTY_RENDER_BUFFERS)");
		goto fail4;
	}

	rc = screen_get_buffer_property_pv(screen_pbuf, SCREEN_PROPERTY_POINTER, &pointer);
	if (rc) {
		perror("screen_get_buffer_property_pv(SCREEN_PROPERTY_POINTER)");
		goto fail4;
	}

	rc = screen_get_buffer_property_iv(screen_pbuf, SCREEN_PROPERTY_STRIDE, &stride);
	if (rc) {
		perror("screen_get_buffer_property_pv(SCREEN_PROPERTY_STRIDE)");
		goto fail4;
	}

	if (interval != 1) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_SWAP_INTERVAL, &interval);
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_SWAP_INTERVAL)");
			goto fail3;
		}
	}

	if( pipeline != -1 ) {
		rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_PIPELINE, &pipeline );
		if (rc) {
			perror("screen_set_window_property_iv(SCREEN_PROPERTY_PIPELINE)");
			goto fail3;
		}
	}

	rc = screen_create_event(&screen_ev);
	if (rc) {
		perror("screen_create_event");
		goto fail4;
	}

	/**
	 ** The animated bar stretches the entire height of the window. This means
	 ** the dirty rects will always be full height, e.g. they will extend from
	 ** 0 to height - 1. We will set that information here instead of repeating
	 ** it each frame.
	 **/

	screen_blit_attr.src_y[1] = rects[1] = 0;
	screen_blit_attr.src_height[1] = rects[3] = size[1];
	screen_blit_attr.dst_y[1] = rects[5] = 0;
	screen_blit_attr.dst_height[1] = rects[7] = size[1];

	/**
	 ** This is our main application loop. It keeps on running unless an error
	 ** occurs. The application loop consists of software rendering calls only.
	 ** We use post_window to handle presentation of our surface.
	 **/

	while (frame_limit_counter < 0 || frame_limit_counter--) {
		while (!screen_get_event(screen_ctx, screen_ev, pause ? ~0 : 0)) {
			rc = screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_TYPE, &val);
			if (rc || val == SCREEN_EVENT_NONE) {
				break;
			}
			switch (val) {
				case SCREEN_EVENT_CLOSE:
					goto end;
				case SCREEN_EVENT_POINTER:
					screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_BUTTONS, &val);
					if (val) {
						screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_POSITION, pos);
						if (pos[0] >= size[0] - exit_area_size &&
							pos[1] < exit_area_size) {
							goto end;
						}
					}
					break;
				case SCREEN_EVENT_KEYBOARD:
					screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_KEY_FLAGS, &val);
					if (val & KEY_DOWN) {
						screen_get_event_property_iv(screen_ev, SCREEN_PROPERTY_KEY_SYM, &val);
						switch (val) {
							case KEYCODE_ESCAPE:
								goto end;
							case KEYCODE_F:
								pause = !pause;
								break;
							default:
								break;
						}
					}
					break;
				case SCREEN_EVENT_MTOUCH_TOUCH:
					rc = screen_set_window_property_iv(screen_win, SCREEN_PROPERTY_ZORDER, &zorder);
					if (rc) {
						perror("screen_set_window_property_iv(SCREEN_PROPERTY_ZORDER)");
						goto end;
					} else {
						printf("setting zorder to %d\n", zorder);
					}
					break;
			}
		}

		if (!pause) {
			switch (format) {
				case SCREEN_FORMAT_RGBA8888:
					draw_rgba8888(pointer, size[0], size[1], stride, 16, 8, 0, 24, rects, bg_rgba[0], bg_rgba[1], bg_rgba[2], bg_rgba[3]);

					if( opt_grid ) {
						draw_grid( screen_ctx, screen_pbuf, size[0], size[1], 16, 8, 0, 24 );
					}
					break;
				case SCREEN_FORMAT_RGBX8888:
					draw_rgba8888(pointer, size[0], size[1], stride, 16, 8, 0, 24, rects, bg_rgba[0], bg_rgba[1], bg_rgba[2], 0x00);
					if( opt_grid ) {
						draw_grid( screen_ctx, screen_pbuf, size[0], size[1], 16, 8, 0, 24 );
					}
					break;
				case SCREEN_FORMAT_RGBA5551:
					draw_rgba5551(pointer, size[0], size[1], stride, 10, 5, 0, 15, rects, bg_rgba[0], bg_rgba[1], bg_rgba[2], bg_rgba[3]);
					if( opt_grid ) {
						draw_grid( screen_ctx, screen_pbuf, size[0], size[1], 10, 5, 0, 15 );
					}
					break;
				case SCREEN_FORMAT_RGBX5551:
					draw_rgba5551(pointer, size[0], size[1], stride, 10, 5, 0, 15, rects, bg_rgba[0], bg_rgba[1], bg_rgba[2], 0x00);
					if( opt_grid ) {
						draw_grid( screen_ctx, screen_pbuf, size[0], size[1], 10, 5, 0, 15 );
					}
					break;
				case SCREEN_FORMAT_RGB565:
					draw_rgb565(pointer, size[0], size[1], stride, 11, 5, 0, rects, bg_rgba[0], bg_rgba[1], bg_rgba[2]);
					if( opt_grid ) {
						draw_grid( screen_ctx, screen_pbuf, size[0], size[1], 11, 5, 0, 24 );
					}
					break;
			}

			rc = screen_get_window_property_pv(screen_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)screen_wbuf);
			if (rc) {
				break;
			}

			screen_blit_attr.src_x[1] = rects[0];
			screen_blit_attr.src_width[1] = rects[2];
			screen_blit_attr.dst_x[1] = rects[0];
			screen_blit_attr.dst_width[1] = rects[2];

			rc = screen_blit(screen_ctx, *screen_wbuf, screen_pbuf, (const int *)&screen_blit_attr);
			if (rc) {
				perror("screen_blit");
				break;
			}

			screen_blit_attr.src_x[1] = rects[4];
			screen_blit_attr.src_width[1] = rects[6];
			screen_blit_attr.dst_x[1] = rects[4];
			screen_blit_attr.dst_width[1] = rects[6];

			rc = screen_blit(screen_ctx, *screen_wbuf, screen_pbuf, (const int *)&screen_blit_attr);
			if (rc) {
				perror("screen_blit");
				break;
			}

			rc = screen_post_window(screen_win, *screen_wbuf, 2, rects, 0);
			if (rc) {
				perror("screen_post_window");
				break;
			}

			/**
			 ** if using -double-buffer, need to fill back
			 ** buffer after post switches front and back buffers
			 **/
			if(nbuffers == 2) {
				rc = screen_get_window_property_pv(screen_win, SCREEN_PROPERTY_RENDER_BUFFERS, (void **)screen_wbuf);
				if (rc) {
					break;
				}

				screen_blit_attr.src_x[1] = rects[0];
				screen_blit_attr.src_width[1] = rects[2];
				screen_blit_attr.dst_x[1] = rects[0];
				screen_blit_attr.dst_width[1] = rects[2];

				rc = screen_blit(screen_ctx, *screen_wbuf, screen_pbuf, (const int *)&screen_blit_attr);
				if (rc) {
					perror("screen_blit");
					break;
				}

				screen_blit_attr.src_x[1] = rects[4];
				screen_blit_attr.src_width[1] = rects[6];
				screen_blit_attr.dst_x[1] = rects[4];
				screen_blit_attr.dst_width[1] = rects[6];

				rc = screen_blit(screen_ctx, *screen_wbuf, screen_pbuf, (const int *)&screen_blit_attr);
				if (rc) {
					perror("screen_blit");
					break;
				}
			}

			{
				static uint64_t last_t;
				static int frames = 0;
				struct timespec to;
				uint64_t t, delta;

				clock_gettime(CLOCK_REALTIME, &to);
				t = timespec2nsec(&to);

				if (frames == 0) {
					last_t = t;
				} else {
					delta = t - last_t;
					if (frames && delta >= 5000000000LL) {
						printf("%d frames in %6.3f seconds = %6.3f FPS\n",
							frames, 0.000000001f * delta,
							1000000000.0f * frames / delta);
						fflush(stdout);
						frames = -1;
					}
				}

				frames++;
			}
		}
	}

end:
	rval = EXIT_SUCCESS;

	screen_destroy_event(screen_ev);
fail4:
	screen_destroy_pixmap(screen_pix);
fail3:
	screen_destroy_window(screen_win);
fail2:
	screen_destroy_context(screen_ctx);
fail1:
	return rval;
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn/product/branches/6.6.0/trunk/apps/screen/tutorials/sw-vsync/sw-vsync.c $ $Rev: 755191 $")
#endif
